from typing import Any, Dict, List, Type, TypeVar

import attr

T = TypeVar("T", bound="GenericDataPoint")


@attr.s(auto_attribs=True)
class GenericDataPoint:
    """ A single data point in a serialised column  """

    x_value: str
    y_value: float
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        x_value = self.x_value
        y_value = self.y_value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "xValue": x_value,
                "yValue": y_value,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        x_value = d.pop("xValue")

        y_value = d.pop("yValue")

        generic_data_point = cls(
            x_value=x_value,
            y_value=y_value,
        )

        generic_data_point.additional_properties = d
        return generic_data_point

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="AlgoExecutionSymbol")


@attr.s(auto_attribs=True)
class AlgoExecutionSymbol:
    """ Any of the reserved python symbols available for use inside an Algo  """

    name: str
    dtype: str
    examples: List[str]
    group: Union[Unset, str] = ""
    description: Union[Unset, str] = ""
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        dtype = self.dtype
        examples = self.examples

        group = self.group
        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "dtype": dtype,
                "examples": examples,
            }
        )
        if group is not UNSET:
            field_dict["group"] = group
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        dtype = d.pop("dtype")

        examples = cast(List[str], d.pop("examples"))

        group = d.pop("group", UNSET)

        description = d.pop("description", UNSET)

        algo_execution_symbol = cls(
            name=name,
            dtype=dtype,
            examples=examples,
            group=group,
            description=description,
        )

        algo_execution_symbol.additional_properties = d
        return algo_execution_symbol

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from mindfoundry.client.horizon.client.horizon_client import Connection, HorizonClient
from mindfoundry.client.horizon.generated.models import (
    IndividualDataset,
    Pipeline,
    PredictResponse,
    ProblemSpecificationConfig,
)
from mindfoundry.client.horizon.generated.models import RegressorType as GeneratedRegressorType
from mindfoundry.client.horizon.generated.models import StageType
from mindfoundry.client.horizon.models.config import (
    CorrelationMethod,
    FeatureGenerationConfig,
    FeatureGeneratorType,
    FilteringConfig,
    ModelConfig,
    RefinementConfig,
    RegressorType,
    StationarizationConfig,
    StationarizationStrategy,
    TargetTransformType,
    assert_not_empty_if_provided,
)

from .base import (
    FeatureSet,
    FlexibleInputData,
    HorizonModel,
    ModelPrediction,
    ReportingCallback,
    ReportingFacility,
)
from .config.base import assert_no_missing_columns


@dataclass
class ForecasterConfig(ModelConfig):
    """ Configures an HorizonForecaster model. """

    target: Optional[str] = None
    """
    Name of the column to be used as target for the forecasting. If not provided in the initial
    config, a target will be required when `.fit()` is invoked.
    Mutually exclusive with `targets` (use one or the other but not both).
    """

    targets: Optional[List[str]] = None
    """
    List of column names to be used as targets for the forecasting. If not provided in the initial
    config, one or more targets will be required when `.fit()` is invoked.
    Mutually exclusive with `target` (use one or the other but not both).
    """

    horizons: Optional[List[int]] = None
    """
    List of horizons to use for training and prediction. A horizon is a number that describes
    how many time steps ahead to predict. One time step is equal to the cadence of the
    data (e.g. daily, hourly, monthly, etc...).
    By default, horizons = [1, 2, 3, 4, 5].
    """

    stationarization: Optional[Union[bool, StationarizationConfig]] = None
    """
    Configuration for the stationarization stage. This is the first stage in the model building
    process. It aims at removing trivial trends in data so as to only leave a stationary process
    which can be more effectively learnt.

    This field can be a boolean that indicates whether the stage should be enabled. It can
    also be a `StationarizationConfig` object in case you need to define finer details about
    the stage itself.

    By default, the stage is enabled.
    """

    feature_generation: Optional[Union[bool, FeatureGenerationConfig]] = None
    """
    Configuration for the feature generation stage. This is the second stage in the model building
    process. It produces artificial features based on existing columns by exploiting a number of
    different augmentation techniques.

    This field can be a boolean that indicates whether the stage should be enabled. It can
    also be a `FeatureGenerationConfig` object in case you need to define finer details about
    the stage itself.

    By default, the stage is enabled.
    """

    filtering: Optional[Union[bool, FilteringConfig]] = None
    """
    Configuration for the filtering stage. This is the third stage in the model building
    process. It reduces the number of total features by computing a relevance score for each
    one and only keeping a selected top N of best performing features.

    This field can be a boolean that indicates whether the stage should be enabled. It can
    also be a `FilteringConfig` object in case you need to define finer details about
    the stage itself.

    By default, the stage is enabled.
    """

    refinement: Optional[Union[bool, RefinementConfig]] = None
    """
    Configuration for the refinement stage. This is the fourth stage in the model building
    process. It further trims down the total number of features by temporarily removing
    specific subsets of features and checking how partial models perform. Those features whose
    removal is the most impactful are assumed to be more important for the prediction.

    This field can be a boolean that indicates whether the stage should be enabled. It can
    also be a `RefinementConfig` object in case you need to define finer details about
    the stage itself.

    By default, the stage is enabled. Beware! this a very computationally intensive stage.
    """

    def validate(self) -> None:
        super().validate()

        assert self.target is None or self.targets is None, (
            "target and targets cannot be specified together at the same time."
            "Use one or the other"
        )

        assert_not_empty_if_provided("target", self.target)
        assert_not_empty_if_provided("targets", self.targets)
        assert_not_empty_if_provided("horizons", self.horizons)

        if isinstance(self.feature_generation, FeatureGenerationConfig):
            self.feature_generation.validate()

        if isinstance(self.filtering, FilteringConfig):
            self.filtering.validate()

        if isinstance(self.stationarization, StationarizationConfig):
            self.stationarization.validate()

        if isinstance(self.refinement, RefinementConfig):
            self.refinement.validate()

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "inputs": self.inputs,
            "exclude_from_input": self.exclude_from_input,
            "target": self.target,
            "targets": self.targets,
            "horizons": self.horizons,
            "stationarization": (
                self.stationarization.to_dict()
                if isinstance(self.stationarization, StationarizationConfig)
                else self.stationarization
            ),
            "feature_generation": (
                self.feature_generation.to_dict()
                if isinstance(self.feature_generation, FeatureGenerationConfig)
                else self.feature_generation
            ),
            "filtering": (
                self.filtering.to_dict()
                if isinstance(self.filtering, FilteringConfig)
                else self.filtering
            ),
            "refinement": (
                self.refinement.to_dict()
                if isinstance(self.refinement, RefinementConfig)
                else self.refinement
            ),
        }

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> ForecasterConfig:
        return ForecasterConfig(
            name=dictionary["name"],
            inputs=dictionary["inputs"],
            exclude_from_input=dictionary["exclude_from_input"],
            target=dictionary["target"],
            targets=dictionary["targets"],
            horizons=dictionary["horizons"],
            stationarization=(
                StationarizationConfig.from_dict(dictionary["stationarization"])
                if isinstance(dictionary["stationarization"], dict)
                else bool(dictionary["stationarization"])
            ),
            feature_generation=(
                FeatureGenerationConfig.from_dict(dictionary["feature_generation"])
                if isinstance(dictionary["feature_generation"], dict)
                else bool(dictionary["feature_generation"])
            ),
            filtering=(
                FilteringConfig.from_dict(dictionary["filtering"])
                if isinstance(dictionary["filtering"], dict)
                else bool(dictionary["filtering"])
            ),
            refinement=(
                RefinementConfig.from_dict(dictionary["refinement"])
                if isinstance(dictionary["refinement"], dict)
                else bool(dictionary["refinement"])
            ),
        )


class HorizonForecaster(HorizonModel):
    def __init__(
        self,
        connection: Connection,
        name: Optional[str] = None,
        target: Optional[str] = None,
        targets: Optional[List[str]] = None,
        inputs: Optional[List[str]] = None,
        exclude_from_input: Optional[List[str]] = None,
        horizons: Optional[List[int]] = None,
        stationarization: Optional[bool] = None,
        stationarization_strategy: Optional[StationarizationStrategy] = None,
        stationarization_target_transform: Optional[TargetTransformType] = None,
        stationarization_adf_threshold: Optional[float] = None,
        feature_generation: Optional[bool] = None,
        feature_generation_max_features: Optional[int] = None,
        feature_generation_generators: Optional[List[FeatureGeneratorType]] = None,
        filtering: Optional[bool] = None,
        filtering_max_features: Optional[int] = None,
        filtering_correlation_method: Optional[CorrelationMethod] = None,
        refinement: Optional[bool] = None,
        refinement_max_features: Optional[int] = None,
        refinement_min_features: Optional[int] = None,
        refinement_early_stopping_sensitivity: Optional[float] = None,
        refinement_deep_search: Optional[bool] = None,
        refinement_regressor: Optional[RegressorType] = None,
    ):
        """
        Create a new Horizon forecasting model.

        :param connection:
            Connection object to the instance you are using, including api key.
        :param name:
            Name of the model. Although not necessary, it is advised to give models a
            name so that multiple models can be easily told apart. If not provided, the model will
            be given a random name.
        :param target:
            Name of the column to be used as target for the forecasting. If not
            provided in the initial config, a target will be required when `.fit()` is invoked.
            Mutually exclusive with `targets` (use one or the other but not both).
        :param targets:
            List of column names to be used as targets for the forecasting.
            If not provided in the initial config, one or more targets will be required when
            `.fit()` is invoked.
            Mutually exclusive with `target` (use one or the other but not both).
        :param inputs:
            List of column names to be used as feature for the forecasting. Only the specified
            columns will be processed and analyzed. Any column that is not in this list will
            be ignored.
            By default, all columns will be used.
            Mutually exclusive with `exclude_from_input` (use one or the other but not both).
        :param exclude_from_input:
            List of column names to be ignored in the forecasting. Every other column that is
            present in the data but not specified in this list will be used as feature.
            By default, no column will be excluded.
            Mutually exclusive with `inputs` (use one or the other but not both).
        :param horizons:
            List of horizons to use for training and prediction. A horizon is a number that
            describe how many time steps ahead in time to predict. One time step is equal to
            the cadence of the data (e.g. daily, hourly, monthly, etc...).
            By default, horizons = [1, 2, 3, 4, 5].
        :param stationarization:
            Whether the stationarization stage is enabled. This is the first stage in the model
            building process. It aims at removing trivial trends in data so as to only leave a
            stationary process which can be more effectively learnt.
            By default, the stage is enabled. If disabled, all other parameters relating to
            stationarization will be ignored.
        :param stationarization_strategy:
            What strategy should be used to handle columns that support / do not support
            stationarization? Default is `keep_fail`.
        :param stationarization_target_transform:
            Whether the target should undergo a lag transformation as well.
            Default is `HorizonLagDiff`.
        :param stationarization_adf_threshold:
            Running stationarisation of features runs an ADF test on all the features. An ADF test
            is a statistical test that assigns a score between 0 and 1 to each feature, where
            scores closer to zero are more likely stationary, and scores closer to 1 are more
            likely non stationary. The score itself is a statistical p-value, and this
            threshold specifies the p-value above which a feature is considered non stationary.
            Default is 0.03.
        :param feature_generation:
            Whether the feature generation stage is enabled. This is the second stage in the model
            building process. It produces artificial features based on existing columns by
            exploiting a number of different augmentation techniques.
            By default, the stage is enabled. If disabled, all other parameters relating to
            feature generation will be ignored.
        :param feature_generation_max_features:
            Upper bound on the number of features to generate per horizon. Default is 500.
        :param feature_generation_generators:
            List of all feature generation strategies to be used in this stage. By default, all
            generators will be used.
        :param filtering:
            Whether the filtering stage is enabled. This is the third stage in the model building
            process. It reduces the number of total features by computing a relevance score for
            each one and only keeping a selected top N of best performing features.
            By default, the stage is enabled. If disabled, all other parameters relating to
            filtering will be ignored.
        :param filtering_max_features:
            Upper bound on the number of remaining features per horizon. Default is 30.
        :param filtering_correlation_method:
            Correlation method to be used to determine the most appropriate features to be
            filtered out and the most useful to instead keep. Default is `pearson`.
        :param refinement:
            Whether the refinement stage is enabled. This is the fourth stage in the model building
            process. It further trims down the total number of features by temporarily removing
            specific subsets of features and checking how partial models perform. Those features
            whose removal is the most impactful are assumed to be more important for the
            prediction.
            By default, the stage is enabled. If disabled, all other parameters relating to
            refinement will be ignored. Beware! this a very computationally intensive stage.
        :param refinement_max_features:
            Upper bound on number of remaining features per horizon. Default is 20.
        :param refinement_min_features:
            Lower bound on the number of remaining features per horizon. Default is 5.
        :param refinement_early_stopping_sensitivity:
            This must be a value between 0 and 1. If closer to 1 then the stage will run faster,
            but will result in a larger feature set. Default is `0`.
        :param refinement_deep_search:
            Whether to try dropping each feature at every iteration, or only a subset (50%).
            Default is `false`.
        :param refinement_regressor:
            Regressor to be used for the refinement. Default is `RandomForest`.
        """
        super().__init__(HorizonClient(connection))
        self._config = ForecasterConfig(
            name=name,
            target=target,
            targets=targets,
            inputs=inputs,
            exclude_from_input=exclude_from_input,
            horizons=horizons,
            stationarization=(
                stationarization
                if stationarization is False
                else StationarizationConfig(
                    strategy=stationarization_strategy,
                    target_transform=stationarization_target_transform,
                    adf_threshold=stationarization_adf_threshold,
                )
            ),
            feature_generation=(
                feature_generation
                if feature_generation is False
                else FeatureGenerationConfig(
                    max_features=feature_generation_max_features,
                    generators=feature_generation_generators,
                )
            ),
            filtering=(
                filtering
                if filtering is False
                else FilteringConfig(
                    max_features=filtering_max_features,
                    correlation_method=filtering_correlation_method,
                )
            ),
            refinement=(
                refinement
                if refinement is False
                else RefinementConfig(
                    max_features=refinement_max_features,
                    min_features=refinement_min_features,
                    early_stopping_sensitivity=refinement_early_stopping_sensitivity,
                    deep_search=refinement_deep_search,
                    regressor=refinement_regressor,
                )
            ),
        )

        self._config.validate()

    @property
    def config(self) -> ForecasterConfig:
        return self._config

    @staticmethod
    def load(file_path: str, connection: Connection) -> HorizonForecaster:
        """
        Load a forecaster model from a file.

        :param file_path:
            Path to a json file that contains a serialized forecaster model. The file must
            have been produced using the `.save()` method.
        :param connection:
            Connection to the Horizon instance to use. Beware that this connection must refer
            to the same Horizon instance that the saved model used.
        """
        with open(file_path, "r+") as file:
            serialized_model = json.loads(file.read())

        assert (
            serialized_model["base_url"] == connection.base_url
        ), "base_url from serialized model does not match base_url from connection"

        # pylint:disable=protected-access
        forecaster = HorizonForecaster(connection)
        forecaster._config = ForecasterConfig.from_dict(serialized_model["config"])
        forecaster._pipeline_id = serialized_model["pipeline_id"]
        forecaster._dataset_id = serialized_model["dataset_id"]

        return forecaster

    def save(self, file_path: str) -> None:
        """
        Save the current forecaster model to a file so it can be loaded at a later time.
        The process persists the following information:
            - configuration of the forecaster;
            - if the `.fit()` method was invoked, a reference to the remote model being built
              after data has been fitted;
            - Horizon instance being used.

        :param file_path:
            Path to a json file where the model will be persisted.
        """
        # pylint:disable=protected-access
        serialized_model = {
            "base_url": self._client._connection.base_url,
            "config": self.config.to_dict(),
            "pipeline_id": self._pipeline_id,
            "dataset_id": self._dataset_id,
        }

        with open(file_path, "w+") as file:
            file.write(json.dumps(serialized_model))

    def fit(
        self,
        training_data: FlexibleInputData,
        target: Optional[str] = None,
        targets: Optional[List[str]] = None,
        reporting_callback: Optional[ReportingCallback] = None,
    ) -> None:
        """
        Create a forecasting model on the given training data. Calling this method a second time
        will delete any previously existing model and create a new one.

        :param training_data:
            A dataframe, series or numpy array. If a series or a one-dimensional numpy array is
            given, Horizon will automatically generate a time index for the time series. We
            anyway recommend to always include a datetime column in your data to be used as
            time index with proper granularity.
            Also notice you should not time-shift your target column when passing it to `.fit`.
            The system will take care of rearranging your data to match the prediction horizon(s).
        :param target:
            Name of the column to use as target. If not specified, the target(s) provided in the
            initial configuration will be used.
            Mutually exclusive with `targets` (use one or the other but not both).
        :param targets:
            Names of the columns to use as targets. If not specified, the target(s) provided in the
            initial configuration will be used.
            Mutually exclusive with `target` (use one or the other but not both).
        :param reporting_callback:
            Function used to report progress of this method. If not provided, a default message
            will be printed to stdout using `print`. Call arguments: current step (int),
            number of total steps (int), description of the current step (str).
        :return:
            The method will wait for the data to be uploaded to Horizon and for the model
            building pipeline to complete. When its coroutine returns, Horizon will have
            completed data fitting.
        """
        reporting_facility = ReportingFacility(
            reporting_callback=reporting_callback,
            steps_count=10,
        )

        reporting_facility.report("Removing pre-existing model")
        self._reset_model()

        input_df = self._make_input_data_frame(training_data)
        dataset_name = self.config.resolve_dataset_name()
        reporting_facility.report(f"Uploading data of shape {input_df.shape} to {dataset_name}")
        dataset = self._client.datasets.upload(df=input_df, name=dataset_name)

        reporting_facility.report(f"Creating new forecasting pipeline for {dataset_name}")
        target_names = self._resolve_targets(dataset, target, targets)
        assert_no_missing_columns(dataset, target_names)

        pipeline = self._client.pipelines.create_custom(
            dataset_id=dataset.summary.id_,
            name=self.config.resolve_pipeline_name(),
        )

        reporting_facility.report(f"Configuring forecasting problem for {target_names}")
        target_features = [str(cp.id_) for cp in dataset.analysis if cp.name in target_names]
        pipeline = self._add_problem_specification_stage(
            pipeline=pipeline,
            dataset=dataset,
            target_features=target_features,
        )

        reporting_facility.report("Configuring stationarization")
        pipeline = self._add_stationarization_stage(pipeline)

        reporting_facility.report("Configuring feature generation")
        pipeline = self._add_feature_generation_stage(pipeline)

        reporting_facility.report("Configuring filtering")
        pipeline = self._add_filtering_stage(pipeline)

        reporting_facility.report("Configuring refinement")
        pipeline = self._add_refinement_stage(pipeline)

        self._dataset_id = dataset.summary.id_
        self._pipeline_id = pipeline.summary.id_

        reporting_facility.report("Running forecasting pipeline")
        self._client.pipelines.run(pipeline.summary.id_, wait_for_completion=True)

        reporting_facility.report("Finalizing forecasting model")
        self._client.pipelines.lock(pipeline.summary.id_)

    def predict(
        self,
        new_data: Optional[FlexibleInputData] = None,
        regressor: Optional[RegressorType] = None,
        target: Optional[str] = None,
        targets: Optional[List[str]] = None,
        reporting_callback: Optional[ReportingCallback] = None,
    ) -> ModelPrediction:
        """
        Predict the future value of target columns using the model obtained by fitting training
        data with the `.fit()` method.

        :param new_data:
            Optional new test data to use. If not provided, the outcome of this invocation will
            be the predicted future values of target columns after the last training data point
            has been processed. If provided, Horizon will feed the new data into the model and
            predict the future values of the targets after the last new data point has been
            processed. Note that this data is not persisted anywhere: it is treated as ephemeral
            data and lost after the prediction is performed.
        :param regressor:
            Regressor to use for modelling. Default is Random Forests.
        :param target:
            Name of the column to use as target. If not specified, the target(s) provided in the
            initial configuration will be used. Must belong to the set of learned targets.
            Mutually exclusive with `targets` (use one or the other but not both).
        :param targets:
            Names of the columns to use as targets. If not specified, the target(s) provided in the
            initial configuration will be used. Must be a subset of learned targets.
            Mutually exclusive with `target` (use one or the other but not both).
        :param reporting_callback:
            Function used to report progress of this method. If not provided, a default message
            will be printed to stdout using `print`. Call arguments: current step (int),
            number of total steps (int), description of the current step (str).
        :return:
            A `ModelPrediction` object that contains all predictions for all horizons (specified
            in the configuration provided in the class constructor) and all target columns.
            For more information on how to interpret results, please read the documentation
            of `ModelPrediction`.
        """
        assert (
            self._pipeline_id is not None and self._dataset_id is not None
        ), "You need to invoke HorizonForecaster.fit() first"

        resolved_horizons = self._resolve_horizons()
        reporting_facility = ReportingFacility(
            reporting_callback=reporting_callback,
            steps_count=2 + len(resolved_horizons),
        )

        reporting_facility.report("Obtaining training information")
        training_dataset = self._client.datasets.get(self._dataset_id)
        resolved_targets = self._resolve_targets(
            dataset=training_dataset,
            target=target,
            targets=targets,
        )
        resolved_new_data = self._make_input_data_frame(new_data) if new_data is not None else None

        prediction_responses_by_horizon: Dict[int, List[PredictResponse]] = {}
        for horizon in resolved_horizons:
            reporting_facility.report(f"Predicting results for horizon = {horizon}")
            prediction_responses_by_horizon[horizon] = self._client.pipelines.predict(
                pipeline_id=self._pipeline_id,
                horizon=horizon,
                new_data=resolved_new_data,
                filter_target_names=resolved_targets,
                regressor=(
                    GeneratedRegressorType(regressor)
                    if regressor is not None
                    else GeneratedRegressorType.RANDOMFOREST
                ),
            )

        reporting_facility.report("Converting predictions into suitable format")
        assert isinstance(training_dataset.summary.cadence, (int, float))
        return ModelPrediction(
            prediction_responses_by_horizon,
            cadence_seconds=float(training_dataset.summary.cadence),
        )

    def update(
        self,
        new_data: FlexibleInputData,
        retrain: Optional[bool] = None,
        reporting_callback: Optional[ReportingCallback] = None,
    ) -> None:
        """
        Append new data onto the existing data specified during `.fit()` (or previous calls
        to this method). Appending data will not cause the model to be learnt again. However,
        subsequent invocations to `.predict()` will take into account any new data.

        :param new_data:
            This data should comply to the following requirements:
                - New data must contain exactly the same columns of the same types
                  w.r.t. existing data currently stored in the dataset;
                - ALL new data points must be at least as recent as the most recent row
                  in the dataset;
                - If the dataset has unique indices, then new data must also have unique indices;
                - New data must have the same cadence as existing data;
                - The oldest data points in the new data must be exactly 0 or 1 cadence away from
                  the most recent row in the dataset.
            If any of the conditions above is violated, the update will fail.
        :param retrain:
            If true, the model will be retrain after new_data has been uploaded.
            Retraining regenerates all features used for prediction.
        :param reporting_callback:
            Function used to report progress of this method. If not provided, a default message
            will be printed to stdout using `print`. Call arguments: current step (int),
            number of total steps (int), description of the current step (str).
        """
        assert (
            self._pipeline_id is not None and self._dataset_id is not None
        ), "You need to invoke HorizonForecaster.fit() first"

        resolved_new_data = self._make_input_data_frame(new_data)
        reporting_facility = ReportingFacility(
            reporting_callback=reporting_callback,
            steps_count=1,
        )

        reporting_facility.report("Uploading new data to Horizon")
        self._client.datasets.append(
            dataset_id=self._dataset_id,
            df=resolved_new_data,
        )

        if retrain is True:
            self._client.pipelines.restart(pipeline_id=self._pipeline_id, wait_for_completion=True)

    def get_features(self) -> FeatureSet:
        """
        Obtain the set of generated features for the current model. Can only be called once
        `.fit` has been invoked.
        """
        assert (
            self._pipeline_id is not None and self._dataset_id is not None
        ), "You need to invoke HorizonForecaster.fit() first"

        horizons = self._resolve_horizons()
        materialized_data: Dict[int, pd.DataFrame] = {
            horizon: self._client.pipelines.download_features(
                pipeline_id=self._pipeline_id,
                horizon=horizon,
                include_targets=True,
            )
            for horizon in horizons
        }

        feature_graph = self._client.pipelines.get_feature_graph(self._pipeline_id)

        return FeatureSet(
            materialized_data=materialized_data,
            nodes_and_links=feature_graph,
        )

    def _add_problem_specification_stage(
        self,
        pipeline: Pipeline,
        dataset: IndividualDataset,
        target_features: List[str],
    ) -> Pipeline:
        return self._add_stage(
            pipeline=pipeline,
            stage_type=StageType.PROBLEM_SPECIFICATION,
            skip=False,
            config=ProblemSpecificationConfig(
                target_features=target_features,
                horizons=self._resolve_horizons(),
                data_split=1,
                active_columns=self.config.resolve_active_columns(dataset),
                scale_factor_multiplier=1,
                used_in_lstm=False,
            ),
        )

    def _add_stationarization_stage(
        self,
        pipeline: Pipeline,
    ) -> Pipeline:
        return self._add_stage(
            pipeline=pipeline,
            stage_type=StageType.STATIONARISATION,
            skip=self.config.stationarization is False,
            config=(
                self.config.stationarization.to_stage_config()
                if isinstance(self.config.stationarization, StationarizationConfig)
                else StationarizationConfig.default_stage_config()
            ),
        )

    def _add_feature_generation_stage(
        self,
        pipeline: Pipeline,
    ) -> Pipeline:
        return self._add_stage(
            pipeline=pipeline,
            stage_type=StageType.FEATURE_GENERATION,
            skip=self.config.feature_generation is False,
            config=(
                self.config.feature_generation.to_stage_config()
                if isinstance(self.config.feature_generation, FeatureGenerationConfig)
                else FeatureGenerationConfig.default_stage_config()
            ),
        )

    def _add_filtering_stage(
        self,
        pipeline: Pipeline,
    ) -> Pipeline:
        return self._add_stage(
            pipeline=pipeline,
            stage_type=StageType.FILTERING,
            skip=self.config.filtering is False,
            config=(
                self.config.filtering.to_stage_config()
                if isinstance(self.config.filtering, FilteringConfig)
                else FilteringConfig.default_stage_config()
            ),
        )

    def _add_refinement_stage(
        self,
        pipeline: Pipeline,
    ) -> Pipeline:
        return self._add_stage(
            pipeline=pipeline,
            stage_type=StageType.REFINEMENT,
            skip=self.config.refinement is False,
            config=(
                self.config.refinement.to_stage_config()
                if isinstance(self.config.refinement, RefinementConfig)
                else RefinementConfig.default_stage_config()
            ),
        )

    def _resolve_targets(
        self,
        dataset: IndividualDataset,
        target: Optional[str] = None,
        targets: Optional[List[str]] = None,
    ) -> List[str]:
        assert target is None or targets is None, (
            "target and targets cannot be specified together at the same time."
            "Use one or the other"
        )

        if target is not None:
            return [target]

        if targets is not None:
            return targets

        if self.config.target is not None:
            return [self.config.target]

        if self.config.targets is not None:
            return self.config.targets

        numeric_columns = [cp.name for cp in dataset.analysis if not cp.is_text]
        assert len(numeric_columns) > 0, "Expected at least one numeric column to use as target"
        return numeric_columns[0:1]

    def _resolve_horizons(self) -> List[int]:
        return self.config.horizons or [1, 2, 3, 4, 5]



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.date_value_pair import DateValuePair

T = TypeVar("T", bound="DateValueSeries")


@attr.s(auto_attribs=True)
class DateValueSeries:
    """ Serialised column, split into contiguous segments  """

    title: str
    data: List[List[DateValuePair]]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        title = self.title
        data = []
        for data_item_data in self.data:
            data_item = []
            for data_item_item_data in data_item_data:
                data_item_item = data_item_item_data.to_dict()

                data_item.append(data_item_item)

            data.append(data_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "title": title,
                "data": data,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        title = d.pop("title")

        data = []
        _data = d.pop("data")
        for data_item_data in _data:
            data_item = []
            _data_item = data_item_data
            for data_item_item_data in _data_item:
                data_item_item = DateValuePair.from_dict(data_item_item_data)

                data_item.append(data_item_item)

            data.append(data_item)

        date_value_series = cls(
            title=title,
            data=data,
        )

        date_value_series.additional_properties = d
        return date_value_series

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
from typing import Any, Dict, Optional

import httpx

from ...client import Client
from ...models.file_upload_form import FileUploadForm
from ...models.ingestion_process import IngestionProcess
from ...types import Response, is_file


def _get_kwargs(
    *,
    client: Client,
    multipart_data: FileUploadForm,
) -> Dict[str, Any]:
    url = "{}/datasets/upload".format(client.base_url)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    files = {}
    data = {}
    for key, value in multipart_data.to_dict().items():
        if is_file(value):
            files[key] = value
        elif isinstance(value, str):
            data[key] = value
        else:
            data[key] = json.dumps(value)

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "files": files,
        "data": data,
    }


def _parse_response(*, response: httpx.Response) -> Optional[IngestionProcess]:
    if response.status_code == 200:
        response_200 = IngestionProcess.from_dict(response.json())

        return response_200
    return None


def _build_response(*, response: httpx.Response) -> Response[IngestionProcess]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    *,
    client: Client,
    multipart_data: FileUploadForm,
) -> Response[IngestionProcess]:
    kwargs = _get_kwargs(
        client=client,
        multipart_data=multipart_data,
    )

    response = httpx.post(
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    *,
    client: Client,
    multipart_data: FileUploadForm,
) -> Optional[IngestionProcess]:
    """An IngestionProcess will be returned, which you can query via the `/ingestionProcess/{id}`
    endpoint until its `status` value is \"completed\". Then take the `datasetId` value and use
    it with the `/datasets/{id}` endpoint to get information on the dataset.

    For example, in python:

        import pandas as pd
        import io
        import json
        import requests
        from time import sleep

        df = pd.DataFrame({\"time\": [\"1/1/1\", \"2/1/1\"], \"x\": [1, 2]})
        str_buffer = io.StringIO(df.to_csv(encoding=\"utf-8\", index=False))
        str_buffer.seek(0)
        str_buffer.name = \"Your Data Set Name\"

        upload_options = {
            \"missingDataStrategy\": {
                \"ffill\": {\"enabled\": True},
                \"replaceMissing\": {\"enabled\": True, \"replaceWith\": 42},
            },
        }

        headers = {\"x-apikey\": \"Paste your API Key here\"}

        base_url = \"https://horizon.mindfoundry.ai\"  # Amend if necessary
        rv = requests.post(
            url=f\"{base_url}/api/datasets/upload\",
            files={\"file\": str_buffer},
            data={\"options\": json.dumps(upload_options)},
            headers=headers,
        )
        ingestion_process = rv.json()
        ingestion_id = ingestion_process[\"id\"]

        while ingestion_process[\"status\"] not in [\"completed\", \"error\"]:
            sleep(0.5)
            rv = requests.get(
                url=f\"{base_url}/api/ingestionProcesses/{ingestion_id}\",
                headers=headers,
            )
            ingestion_process = rv.json()

        if ingestion_process[\"status\"] == \"completed\":
            dataset_id = ingestion_process[\"datasetId\"]
            rv = requests.get(url=f\"{base_url}/api/datasets/{dataset_id}\", headers=headers)
            dataset = rv.json()
        else:
            raise Exception(ingestion_process[\"error\"])
    """

    return sync_detailed(
        client=client,
        multipart_data=multipart_data,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    multipart_data: FileUploadForm,
) -> Response[IngestionProcess]:
    kwargs = _get_kwargs(
        client=client,
        multipart_data=multipart_data,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.post(**kwargs)

    return _build_response(response=response)


async def asyncio(
    *,
    client: Client,
    multipart_data: FileUploadForm,
) -> Optional[IngestionProcess]:
    """An IngestionProcess will be returned, which you can query via the `/ingestionProcess/{id}`
    endpoint until its `status` value is \"completed\". Then take the `datasetId` value and use
    it with the `/datasets/{id}` endpoint to get information on the dataset.

    For example, in python:

        import pandas as pd
        import io
        import json
        import requests
        from time import sleep

        df = pd.DataFrame({\"time\": [\"1/1/1\", \"2/1/1\"], \"x\": [1, 2]})
        str_buffer = io.StringIO(df.to_csv(encoding=\"utf-8\", index=False))
        str_buffer.seek(0)
        str_buffer.name = \"Your Data Set Name\"

        upload_options = {
            \"missingDataStrategy\": {
                \"ffill\": {\"enabled\": True},
                \"replaceMissing\": {\"enabled\": True, \"replaceWith\": 42},
            },
        }

        headers = {\"x-apikey\": \"Paste your API Key here\"}

        base_url = \"https://horizon.mindfoundry.ai\"  # Amend if necessary
        rv = requests.post(
            url=f\"{base_url}/api/datasets/upload\",
            files={\"file\": str_buffer},
            data={\"options\": json.dumps(upload_options)},
            headers=headers,
        )
        ingestion_process = rv.json()
        ingestion_id = ingestion_process[\"id\"]

        while ingestion_process[\"status\"] not in [\"completed\", \"error\"]:
            sleep(0.5)
            rv = requests.get(
                url=f\"{base_url}/api/ingestionProcesses/{ingestion_id}\",
                headers=headers,
            )
            ingestion_process = rv.json()

        if ingestion_process[\"status\"] == \"completed\":
            dataset_id = ingestion_process[\"datasetId\"]
            rv = requests.get(url=f\"{base_url}/api/datasets/{dataset_id}\", headers=headers)
            dataset = rv.json()
        else:
            raise Exception(ingestion_process[\"error\"])
    """

    return (
        await asyncio_detailed(
            client=client,
            multipart_data=multipart_data,
        )
    ).parsed



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.backtest_stage import BacktestStage
from ..models.classification_backtest_stage import ClassificationBacktestStage
from ..models.classification_discovery_stage import ClassificationDiscoveryStage
from ..models.classification_specification_stage import ClassificationSpecificationStage
from ..models.dataset import Dataset
from ..models.feature_generation_stage import FeatureGenerationStage
from ..models.filter_stage import FilterStage
from ..models.lstm_backtest_stage import LstmBacktestStage
from ..models.lstm_prediction_stage import LstmPredictionStage
from ..models.pipeline_summary import PipelineSummary
from ..models.prediction_stage import PredictionStage
from ..models.problem_specification_stage import ProblemSpecificationStage
from ..models.refinement_stage import RefinementStage
from ..models.stationarisation_stage import StationarisationStage
from ..models.trading_simulation_stage import TradingSimulationStage
from ..models.trading_specification_stage import TradingSpecificationStage

T = TypeVar("T", bound="Pipeline")


@attr.s(auto_attribs=True)
class Pipeline:
    """A Pipeline consists of a sequence of stages, which can be run to perform analysis,
    generate features, or make predictions from a dataset.

    Each stage produces an output, which is fed to the next stage.
    It also produces Insights, which can be retrieved separately via the API.
    Pipelines are linear. Stages within the list are in order."""

    summary: PipelineSummary
    stages: List[
        Union[
            BacktestStage,
            ClassificationBacktestStage,
            ClassificationDiscoveryStage,
            ClassificationSpecificationStage,
            FeatureGenerationStage,
            FilterStage,
            LstmBacktestStage,
            LstmPredictionStage,
            PredictionStage,
            ProblemSpecificationStage,
            RefinementStage,
            StationarisationStage,
            TradingSimulationStage,
            TradingSpecificationStage,
        ]
    ]
    dataset: Dataset
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        summary = self.summary.to_dict()

        stages = []
        for stages_item_data in self.stages:
            if isinstance(stages_item_data, ProblemSpecificationStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, StationarisationStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, FeatureGenerationStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, FilterStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, RefinementStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, BacktestStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, PredictionStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, TradingSpecificationStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, TradingSimulationStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, ClassificationSpecificationStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, ClassificationDiscoveryStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, ClassificationBacktestStage):
                stages_item = stages_item_data.to_dict()

            elif isinstance(stages_item_data, LstmBacktestStage):
                stages_item = stages_item_data.to_dict()

            else:
                stages_item = stages_item_data.to_dict()

            stages.append(stages_item)

        dataset = self.dataset.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "summary": summary,
                "stages": stages,
                "dataset": dataset,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        summary = PipelineSummary.from_dict(d.pop("summary"))

        stages = []
        _stages = d.pop("stages")
        for stages_item_data in _stages:

            def _parse_stages_item(
                data: object,
            ) -> Union[
                BacktestStage,
                ClassificationBacktestStage,
                ClassificationDiscoveryStage,
                ClassificationSpecificationStage,
                FeatureGenerationStage,
                FilterStage,
                LstmBacktestStage,
                LstmPredictionStage,
                PredictionStage,
                ProblemSpecificationStage,
                RefinementStage,
                StationarisationStage,
                TradingSimulationStage,
                TradingSpecificationStage,
            ]:
                try:
                    stages_item_type0: ProblemSpecificationStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type0 = ProblemSpecificationStage.from_dict(data)

                    return stages_item_type0
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type1: StationarisationStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type1 = StationarisationStage.from_dict(data)

                    return stages_item_type1
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type2: FeatureGenerationStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type2 = FeatureGenerationStage.from_dict(data)

                    return stages_item_type2
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type3: FilterStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type3 = FilterStage.from_dict(data)

                    return stages_item_type3
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type4: RefinementStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type4 = RefinementStage.from_dict(data)

                    return stages_item_type4
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type5: BacktestStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type5 = BacktestStage.from_dict(data)

                    return stages_item_type5
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type6: PredictionStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type6 = PredictionStage.from_dict(data)

                    return stages_item_type6
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type7: TradingSpecificationStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type7 = TradingSpecificationStage.from_dict(data)

                    return stages_item_type7
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type8: TradingSimulationStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type8 = TradingSimulationStage.from_dict(data)

                    return stages_item_type8
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type9: ClassificationSpecificationStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type9 = ClassificationSpecificationStage.from_dict(data)

                    return stages_item_type9
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type10: ClassificationDiscoveryStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type10 = ClassificationDiscoveryStage.from_dict(data)

                    return stages_item_type10
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type11: ClassificationBacktestStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type11 = ClassificationBacktestStage.from_dict(data)

                    return stages_item_type11
                except:  # noqa: E722
                    pass
                try:
                    stages_item_type12: LstmBacktestStage
                    if not isinstance(data, dict):
                        raise TypeError()
                    stages_item_type12 = LstmBacktestStage.from_dict(data)

                    return stages_item_type12
                except:  # noqa: E722
                    pass
                if not isinstance(data, dict):
                    raise TypeError()
                stages_item_type13: LstmPredictionStage
                stages_item_type13 = LstmPredictionStage.from_dict(data)

                return stages_item_type13

            stages_item = _parse_stages_item(stages_item_data)

            stages.append(stages_item)

        dataset = Dataset.from_dict(d.pop("dataset"))

        pipeline = cls(
            summary=summary,
            stages=stages,
            dataset=dataset,
        )

        pipeline.additional_properties = d
        return pipeline

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from mindfoundry.client.horizon.client.horizon_client import Connection, HorizonClient
from mindfoundry.client.horizon.generated.models import (
    ClassificationSpecificationConfig,
    IndividualDataset,
    Pipeline,
    StageType,
)
from mindfoundry.client.horizon.models.config import (
    ClassificationDiscoveryConfig,
    ModelConfig,
    assert_not_empty_if_provided,
)

from .base import FlexibleInputData, HorizonModel, ReportingCallback, ReportingFacility


@dataclass
class ClassifierConfig(ModelConfig):
    target: Optional[str] = None
    """
    Name of the column to be used as target for the classification. If not provided in the initial
    config, a target will be required when `.fit()` is invoked.
    """

    label_to_predict: Optional[str] = None
    """
    Single value belonging to the target column domain. This indicates what the classifier will
    attempt to predict. The default is the first value available in the training data.
    """

    discovery: Optional[ClassificationDiscoveryConfig] = None
    """
    Configuration for the discovery stage. This is the first stage in the model building
    process. It produces artificial features based on existing columns by exploiting a number of
    different augmentation techniques.

    This field can be a boolean that indicates whether the stage should be enabled. It can
    also be a `ClassificationDiscoveryConfig` object in case you need to define finer details about
    the stage itself.

    By default, the stage is enabled.
    """

    def validate(self) -> None:
        super().validate()

        assert_not_empty_if_provided("target", self.target)
        assert_not_empty_if_provided("label_to_predict", self.label_to_predict)

        if isinstance(self.discovery, ClassificationDiscoveryConfig):
            self.discovery.validate()

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "inputs": self.inputs,
            "exclude_from_input": self.exclude_from_input,
            "target": self.target,
            "label_to_predict": self.label_to_predict,
            "discovery": (
                self.discovery.to_dict()
                if isinstance(self.discovery, ClassificationDiscoveryConfig)
                else self.discovery
            ),
        }

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> ClassifierConfig:
        return ClassifierConfig(
            name=dictionary["name"],
            inputs=dictionary["inputs"],
            exclude_from_input=dictionary["exclude_from_input"],
            target=dictionary["target"],
            label_to_predict=dictionary["label_to_predict"],
            discovery=(
                ClassificationDiscoveryConfig.from_dict(dictionary["discovery"])
                if isinstance(dictionary["discovery"], dict)
                else dictionary["discovery"]
            ),
        )


class HorizonClassifier(HorizonModel):
    def __init__(
        self,
        connection: Connection,
        name: Optional[str] = None,
        target: Optional[str] = None,
        inputs: Optional[List[str]] = None,
        exclude_from_input: Optional[List[str]] = None,
        label_to_predict: Optional[str] = None,
        discovery_max_features: Optional[int] = None,
        discovery_timeout_seconds: Optional[float] = None,
        discovery_excluded_columns: Optional[List[str]] = None,
        discovery_feature_generation_enabled: Optional[bool] = None,
    ):
        """
        Create a new Horizon classification model.
        :param connection:
            Connection object to the instance you are using, including api key.
        :param name:
            Name of the model. Although not necessary, it is advised to give models a
            name so that multiple models can be easily told apart. If not provided, the model will
            be given a random name.
        :param target:
            Name of the column to be used as target for the classification. If not
            provided in the initial config, a target will be required when `.fit()` is invoked.
        :param inputs:
            List of column names to be used as feature for the forecasting. Only the specified
            columns will be processed and analyzed. Any column that is not in this list will
            be ignored.
            By default, all columns will be used.
            Mutually exclusive with `exclude_from_input` (use one or the other but not both).
        :param exclude_from_input:
            List of column names to be ignored in the forecasting. Every other column that is
            present in the data but not specified in this list will be used as feature.
            By default, no column will be excluded.
            Mutually exclusive with `inputs` (use one or the other but not both).
        :param discovery_max_features:
            Upper bound on the number of features to generate per horizon. Default is 100.
        :param discovery_timeout_seconds:
            Maximum time allowed for running this stage. Default is `60` seconds.
        :param discovery_excluded_columns:
            List of column names to exclude from feature generation. This is the first stage
            in the model building process. It produces artificial features based on existing
            columns by exploiting a number of different augmentation techniques.
            By default, no column will be excluded.
        :param discovery_feature_generation_enabled:
            Discovery can run either with or without feature generation. In the latter case,
            only the original columns will be taken into account and Horizon will not generate
            any artificial feature. Default is True.

        """
        super().__init__(HorizonClient(connection))

        self._config = ClassifierConfig(
            name=name,
            inputs=inputs,
            exclude_from_input=exclude_from_input,
            target=target,
            label_to_predict=label_to_predict,
            discovery=ClassificationDiscoveryConfig(
                max_features=discovery_max_features,
                timeout_seconds=discovery_timeout_seconds,
                excluded_columns=discovery_excluded_columns,
                feature_generation_enabled=discovery_feature_generation_enabled,
            ),
        )

        self._config.validate()

    @property
    def config(self) -> ClassifierConfig:
        return self._config

    @staticmethod
    def load(file_path: str, connection: Connection) -> HorizonClassifier:
        """
        Load a classifier model from a file.

        :param file_path:
            Path to a json file that contains a serialized classifier model. The file must
            have been produced using the `.save()` method.
        :param connection:
            Connection to the Horizon instance to use. Beware that this connection must refer
            to the same Horizon instance that the saved model used.
        """
        with open(file_path, "r+") as file:
            serialized_model = json.loads(file.read())

        assert (
            serialized_model["base_url"] == connection.base_url
        ), "base_url from serialized model does not match base_url from connection"

        # pylint:disable=protected-access
        forecaster = HorizonClassifier(connection)
        forecaster._config = ClassifierConfig.from_dict(serialized_model["config"])
        forecaster._pipeline_id = serialized_model["pipeline_id"]
        forecaster._dataset_id = serialized_model["dataset_id"]

        return forecaster

    def save(self, file_path: str) -> None:
        """
        Save the current classifier model to a file so it can be loaded at a later time.
        The process persists the following information:
            - configuration of the classifier;
            - if the `.fit()` method was invoked, a reference to the remote model being built
              after data has been fitted;
            - Horizon instance being used.

        :param file_path:
            Path to a json file where the model will be persisted.
        """
        # pylint:disable=protected-access
        serialized_model = {
            "base_url": self._client._connection.base_url,
            "config": self.config.to_dict(),
            "pipeline_id": self._pipeline_id,
            "dataset_id": self._dataset_id,
        }

        with open(file_path, "w+") as file:
            file.write(json.dumps(serialized_model))

    def fit(
        self,
        x: FlexibleInputData,
        y: Union[pd.Series, np.ndarray],
        reporting_callback: Optional[ReportingCallback] = None,
    ) -> None:
        """
        Create a classification model on the given training data. Calling this method a second time
        will delete any previously existing model and create a new one.

        :param x:
            Features to be used for classification.
            A dataframe, series or numpy array. If a series or a one-dimensional numpy array is
            given, Horizon will automatically generate a time index for the time series. We
            anyway recommend to always include a datetime column in your data to be used as
            time index with proper granularity.
        :param y:
            Class to be used for classification. Must be of the same length as x.
            Can only contain two distinct values. Any more will result in an error.
        :param reporting_callback:
            Function used to report progress of this method. If not provided, a default message
            will be printed to stdout using `print`. Call arguments: current step (int),
            number of total steps (int), description of the current step (str).
        :return:
            The method will wait for the data to be uploaded to Horizon and for the model
            building pipeline to complete. When its coroutine returns, Horizon will have
            completed data fitting.
        """
        reporting_facility = ReportingFacility(
            reporting_callback=reporting_callback,
            steps_count=7,
        )

        reporting_facility.report("Removing pre-existing model")
        self._reset_model()

        training_data = self._make_input_data_frame(x)
        target_name = self._find_name_for_target_column(training_data)
        training_data[target_name] = y

        dataset_name = self.config.resolve_dataset_name()
        reporting_facility.report(
            f"Uploading data of shape {training_data.shape} to {dataset_name}"
        )
        dataset = self._client.datasets.upload(df=training_data, name=dataset_name)

        reporting_facility.report(f"Creating new classification pipeline for {dataset_name}")
        pipeline = self._client.pipelines.create_custom(
            dataset_id=dataset.summary.id_,
            name=self.config.resolve_pipeline_name(),
        )

        reporting_facility.report(f"Configuring classification problem for {target_name}")
        pipeline = self._add_problem_specification_stage(pipeline, dataset, target_name)

        reporting_facility.report("Configuring feature discovery")
        pipeline = self._add_discovery_stage(pipeline)

        self._dataset_id = dataset.summary.id_
        self._pipeline_id = pipeline.summary.id_

        reporting_facility.report("Running classification pipeline")
        self._client.pipelines.run(pipeline.summary.id_, wait_for_completion=True)

        reporting_facility.report("Finalizing classification model")
        self._client.pipelines.lock(pipeline.summary.id_)

    def update(
        self,
        new_x: FlexibleInputData,
        new_y: Union[pd.Series, np.ndarray],
        reporting_callback: Optional[ReportingCallback] = None,
    ) -> None:
        """
        Append new data onto the existing data specified during `.fit()` (or previous calls
        to this method). Appending data will not cause the model to be learnt again. However,
        subsequent invocations to `.predict()` will take into account any new data.

        :param new_x:
            This data should comply to the following requirements:
                - New data must contain exactly the same columns of the same types
                  w.r.t. existing data currently stored in the dataset;
                - ALL new data points must be at least as recent as the most recent row
                  in the dataset;
                - If the dataset has unique indices, then new data must also have unique indices;
                - New data must have the same cadence as existing data;
                - The oldest data points in the new data must be exactly 0 or 1 cadence away from
                  the most recent row in the dataset.
            If any of the conditions above is violated, the update will fail.
        :param reporting_callback:
            Function used to report progress of this method. If not provided, a default message
            will be printed to stdout using `print`. Call arguments: current step (int),
            number of total steps (int), description of the current step (str).
        """
        assert (
            self._pipeline_id is not None and self._dataset_id is not None
        ), "You need to invoke HorizonClassifier.fit() first"

        resolved_new_data = self._make_input_data_frame(new_x)
        target_name = self._find_name_for_target_column(resolved_new_data)
        resolved_new_data[target_name] = new_y

        reporting_facility = ReportingFacility(
            reporting_callback=reporting_callback,
            steps_count=1,
        )

        reporting_facility.report("Uploading new data to Horizon")
        self._client.datasets.append(
            dataset_id=self._dataset_id,
            df=resolved_new_data,
        )

    def _add_problem_specification_stage(
        self,
        pipeline: Pipeline,
        dataset: IndividualDataset,
        target_name: str,
    ) -> Pipeline:
        matching_target_ids = [cp.id_ for cp in dataset.analysis if cp.name == target_name]
        assert len(matching_target_ids) == 1, "Dataset must contain one and only one target column"

        if self.config.label_to_predict is not None:
            label_to_predict = self.config.label_to_predict
        else:
            target = next(cp for cp in dataset.analysis if cp.name == target_name)
            assert isinstance(target.binary_labels, list)
            label_to_predict = target.binary_labels[0]

        return self._add_stage(
            pipeline=pipeline,
            stage_type=StageType.CLASSIFICATION_SPECIFICATION,
            skip=False,
            config=ClassificationSpecificationConfig(
                target_id=str(matching_target_ids[0]),
                last_train_timestamp_millisec=None,
                label_to_predict=label_to_predict,
                active_columns=self.config.resolve_active_columns(dataset),
            ),
        )

    def _add_discovery_stage(self, pipeline: Pipeline) -> Pipeline:
        return self._add_stage(
            pipeline=pipeline,
            stage_type=StageType.CLASSIFICATION_DISCOVERY,
            skip=False,
            config=(
                self.config.discovery.to_stage_config()
                if self.config.discovery is not None
                else ClassificationDiscoveryConfig.default_stage_config()
            ),
        )

    @staticmethod
    def _find_name_for_target_column(training_data: pd.DataFrame) -> str:
        return next(
            f"target_{i}" for i in range(0, 100) if f"target_{i}" not in set(training_data.columns)
        )



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..models.column_passport_autocorrelations import ColumnPassportAutocorrelations
from ..types import UNSET, Unset

T = TypeVar("T", bound="ColumnPassport")


@attr.s(auto_attribs=True)
class ColumnPassport:
    """ Summary statistics / properties of a raw column. Calculated on upload of a dataset. """

    id_: int
    name: str
    cadence: float
    n_rows: int
    autocorrelations: ColumnPassportAutocorrelations
    is_text: Union[Unset, bool] = False
    is_binary: Union[Unset, bool] = False
    binary_labels: Union[Unset, List[str]] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id_ = self.id_
        name = self.name
        cadence = self.cadence
        n_rows = self.n_rows
        autocorrelations = self.autocorrelations.to_dict()

        is_text = self.is_text
        is_binary = self.is_binary
        binary_labels: Union[Unset, List[str]] = UNSET
        if not isinstance(self.binary_labels, Unset):
            binary_labels = self.binary_labels

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id_,
                "name": name,
                "cadence": cadence,
                "nRows": n_rows,
                "autocorrelations": autocorrelations,
            }
        )
        if is_text is not UNSET:
            field_dict["isText"] = is_text
        if is_binary is not UNSET:
            field_dict["isBinary"] = is_binary
        if binary_labels is not UNSET:
            field_dict["binaryLabels"] = binary_labels

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id_ = d.pop("id")

        name = d.pop("name")

        cadence = d.pop("cadence")

        n_rows = d.pop("nRows")

        autocorrelations = ColumnPassportAutocorrelations.from_dict(d.pop("autocorrelations"))

        is_text = d.pop("isText", UNSET)

        is_binary = d.pop("isBinary", UNSET)

        binary_labels = cast(List[str], d.pop("binaryLabels", UNSET))

        column_passport = cls(
            id_=id_,
            name=name,
            cadence=cadence,
            n_rows=n_rows,
            autocorrelations=autocorrelations,
            is_text=is_text,
            is_binary=is_binary,
            binary_labels=binary_labels,
        )

        column_passport.additional_properties = d
        return column_passport

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union

import attr
from dateutil.parser import isoparse

from ..models.raw_column import RawColumn
from ..models.storage_specification import StorageSpecification
from ..types import UNSET, Unset

T = TypeVar("T", bound="Dataset")


@attr.s(auto_attribs=True)
class Dataset:
    """  """

    name: str
    id_: int
    columns: List[RawColumn]
    n_rows: int
    n_columns: int
    upload_date: datetime.datetime
    upload_user_id: int
    time_index: str
    is_time_index_generated: bool
    storage_specification: StorageSpecification
    source_file_bytes: int
    last_update_date: Union[Unset, None, datetime.datetime] = UNSET
    last_update_error: Union[Unset, None, str] = UNSET
    description: Union[Unset, str] = "Description unspecified"
    storage_size_bytes: Union[Unset, int] = 0
    first_index_timestamp: Union[Unset, None, int] = UNSET
    last_index_timestamp: Union[Unset, None, int] = UNSET
    indices_unique: Union[Unset, bool] = False
    cadence: Union[Unset, None, float] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        id_ = self.id_
        columns = []
        for columns_item_data in self.columns:
            columns_item = columns_item_data.to_dict()

            columns.append(columns_item)

        n_rows = self.n_rows
        n_columns = self.n_columns
        upload_date = self.upload_date.isoformat()

        upload_user_id = self.upload_user_id
        time_index = self.time_index
        is_time_index_generated = self.is_time_index_generated
        storage_specification = self.storage_specification.to_dict()

        source_file_bytes = self.source_file_bytes
        last_update_date: Union[Unset, None, str] = UNSET
        if not isinstance(self.last_update_date, Unset):
            last_update_date = self.last_update_date.isoformat() if self.last_update_date else None

        last_update_error = self.last_update_error
        description = self.description
        storage_size_bytes = self.storage_size_bytes
        first_index_timestamp = self.first_index_timestamp
        last_index_timestamp = self.last_index_timestamp
        indices_unique = self.indices_unique
        cadence = self.cadence

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "id": id_,
                "columns": columns,
                "nRows": n_rows,
                "nColumns": n_columns,
                "uploadDate": upload_date,
                "uploadUserId": upload_user_id,
                "timeIndex": time_index,
                "isTimeIndexGenerated": is_time_index_generated,
                "storageSpecification": storage_specification,
                "sourceFileBytes": source_file_bytes,
            }
        )
        if last_update_date is not UNSET:
            field_dict["lastUpdateDate"] = last_update_date
        if last_update_error is not UNSET:
            field_dict["lastUpdateError"] = last_update_error
        if description is not UNSET:
            field_dict["description"] = description
        if storage_size_bytes is not UNSET:
            field_dict["storageSizeBytes"] = storage_size_bytes
        if first_index_timestamp is not UNSET:
            field_dict["firstIndexTimestamp"] = first_index_timestamp
        if last_index_timestamp is not UNSET:
            field_dict["lastIndexTimestamp"] = last_index_timestamp
        if indices_unique is not UNSET:
            field_dict["indicesUnique"] = indices_unique
        if cadence is not UNSET:
            field_dict["cadence"] = cadence

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        id_ = d.pop("id")

        columns = []
        _columns = d.pop("columns")
        for columns_item_data in _columns:
            columns_item = RawColumn.from_dict(columns_item_data)

            columns.append(columns_item)

        n_rows = d.pop("nRows")

        n_columns = d.pop("nColumns")

        upload_date = isoparse(d.pop("uploadDate"))

        upload_user_id = d.pop("uploadUserId")

        time_index = d.pop("timeIndex")

        is_time_index_generated = d.pop("isTimeIndexGenerated")

        storage_specification = StorageSpecification.from_dict(d.pop("storageSpecification"))

        source_file_bytes = d.pop("sourceFileBytes")

        last_update_date = None
        _last_update_date = d.pop("lastUpdateDate", UNSET)
        if _last_update_date is not None and not isinstance(_last_update_date, Unset):
            last_update_date = isoparse(_last_update_date)

        last_update_error = d.pop("lastUpdateError", UNSET)

        description = d.pop("description", UNSET)

        storage_size_bytes = d.pop("storageSizeBytes", UNSET)

        first_index_timestamp = d.pop("firstIndexTimestamp", UNSET)

        last_index_timestamp = d.pop("lastIndexTimestamp", UNSET)

        indices_unique = d.pop("indicesUnique", UNSET)

        cadence = d.pop("cadence", UNSET)

        dataset = cls(
            name=name,
            id_=id_,
            columns=columns,
            n_rows=n_rows,
            n_columns=n_columns,
            upload_date=upload_date,
            upload_user_id=upload_user_id,
            time_index=time_index,
            is_time_index_generated=is_time_index_generated,
            storage_specification=storage_specification,
            source_file_bytes=source_file_bytes,
            last_update_date=last_update_date,
            last_update_error=last_update_error,
            description=description,
            storage_size_bytes=storage_size_bytes,
            first_index_timestamp=first_index_timestamp,
            last_index_timestamp=last_index_timestamp,
            indices_unique=indices_unique,
            cadence=cadence,
        )

        dataset.additional_properties = d
        return dataset

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class FeatureGeneratorType(str, Enum):
    AUTOLAG = "autolag"
    EWMA = "ewma"
    LAG = "lag"
    LOGARITHM = "logarithm"
    ROLLING_AVERAGE = "rolling_average"
    CALENDAR = "calendar"
    PERC_CHANGE = "perc_change"
    NUM_PEAKS = "num_peaks"
    ONE_HOT_ENCODE = "one_hot_encode"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.date_value_point_schema import DateValuePointSchema
from ..models.label_frequencies_grouped_by_time import LabelFrequenciesGroupedByTime
from ..models.ohlc_point_schema import OHLCPointSchema
from ..models.text_label_counts import TextLabelCounts
from ..types import UNSET, Unset

T = TypeVar("T", bound="SeriesOverviewSchema")


@attr.s(auto_attribs=True)
class SeriesOverviewSchema:
    """  """

    series_name: str
    ohlc_data: List[List[OHLCPointSchema]]
    close_data: List[List[DateValuePointSchema]]
    intraday_available: bool
    text_highest_frequencies: List[TextLabelCounts]
    frequencies_grouped_by_time: Union[Unset, LabelFrequenciesGroupedByTime] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        series_name = self.series_name
        ohlc_data = []
        for ohlc_data_item_data in self.ohlc_data:
            ohlc_data_item = []
            for ohlc_data_item_item_data in ohlc_data_item_data:
                ohlc_data_item_item = ohlc_data_item_item_data.to_dict()

                ohlc_data_item.append(ohlc_data_item_item)

            ohlc_data.append(ohlc_data_item)

        close_data = []
        for close_data_item_data in self.close_data:
            close_data_item = []
            for close_data_item_item_data in close_data_item_data:
                close_data_item_item = close_data_item_item_data.to_dict()

                close_data_item.append(close_data_item_item)

            close_data.append(close_data_item)

        intraday_available = self.intraday_available
        text_highest_frequencies = []
        for text_highest_frequencies_item_data in self.text_highest_frequencies:
            text_highest_frequencies_item = text_highest_frequencies_item_data.to_dict()

            text_highest_frequencies.append(text_highest_frequencies_item)

        frequencies_grouped_by_time: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.frequencies_grouped_by_time, Unset):
            frequencies_grouped_by_time = self.frequencies_grouped_by_time.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "seriesName": series_name,
                "ohlcData": ohlc_data,
                "closeData": close_data,
                "intradayAvailable": intraday_available,
                "textHighestFrequencies": text_highest_frequencies,
            }
        )
        if frequencies_grouped_by_time is not UNSET:
            field_dict["frequenciesGroupedByTime"] = frequencies_grouped_by_time

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        series_name = d.pop("seriesName")

        ohlc_data = []
        _ohlc_data = d.pop("ohlcData")
        for ohlc_data_item_data in _ohlc_data:
            ohlc_data_item = []
            _ohlc_data_item = ohlc_data_item_data
            for ohlc_data_item_item_data in _ohlc_data_item:
                ohlc_data_item_item = OHLCPointSchema.from_dict(ohlc_data_item_item_data)

                ohlc_data_item.append(ohlc_data_item_item)

            ohlc_data.append(ohlc_data_item)

        close_data = []
        _close_data = d.pop("closeData")
        for close_data_item_data in _close_data:
            close_data_item = []
            _close_data_item = close_data_item_data
            for close_data_item_item_data in _close_data_item:
                close_data_item_item = DateValuePointSchema.from_dict(close_data_item_item_data)

                close_data_item.append(close_data_item_item)

            close_data.append(close_data_item)

        intraday_available = d.pop("intradayAvailable")

        text_highest_frequencies = []
        _text_highest_frequencies = d.pop("textHighestFrequencies")
        for text_highest_frequencies_item_data in _text_highest_frequencies:
            text_highest_frequencies_item = TextLabelCounts.from_dict(
                text_highest_frequencies_item_data
            )

            text_highest_frequencies.append(text_highest_frequencies_item)

        frequencies_grouped_by_time: Union[Unset, LabelFrequenciesGroupedByTime] = UNSET
        _frequencies_grouped_by_time = d.pop("frequenciesGroupedByTime", UNSET)
        if not isinstance(_frequencies_grouped_by_time, Unset):
            frequencies_grouped_by_time = LabelFrequenciesGroupedByTime.from_dict(
                _frequencies_grouped_by_time
            )

        series_overview_schema = cls(
            series_name=series_name,
            ohlc_data=ohlc_data,
            close_data=close_data,
            intraday_available=intraday_available,
            text_highest_frequencies=text_highest_frequencies,
            frequencies_grouped_by_time=frequencies_grouped_by_time,
        )

        series_overview_schema.additional_properties = d
        return series_overview_schema

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
from typing import Any, Dict, Optional

import httpx

from ...client import Client
from ...models.classification_prediction_form import ClassificationPredictionForm
from ...models.classification_prediction_result import ClassificationPredictionResult
from ...types import Response, is_file


def _get_kwargs(
    *,
    client: Client,
    id_: int,
    multipart_data: ClassificationPredictionForm,
) -> Dict[str, Any]:
    url = "{}/pipelines/{id}/expert/classification/predict".format(client.base_url, id=id_)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    files = {}
    data = {}
    for key, value in multipart_data.to_dict().items():
        if is_file(value):
            files[key] = value
        elif isinstance(value, str):
            data[key] = value
        else:
            data[key] = json.dumps(value)

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "files": files,
        "data": data,
    }


def _parse_response(*, response: httpx.Response) -> Optional[ClassificationPredictionResult]:
    if response.status_code == 200:
        response_200 = ClassificationPredictionResult.from_dict(response.json())

        return response_200
    return None


def _build_response(*, response: httpx.Response) -> Response[ClassificationPredictionResult]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    *,
    client: Client,
    id_: int,
    multipart_data: ClassificationPredictionForm,
) -> Response[ClassificationPredictionResult]:
    kwargs = _get_kwargs(
        client=client,
        id_=id_,
        multipart_data=multipart_data,
    )

    response = httpx.post(
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    *,
    client: Client,
    id_: int,
    multipart_data: ClassificationPredictionForm,
) -> Optional[ClassificationPredictionResult]:
    """It uses features inferred from on the original pipeline dataset, but a new
    model is trained on the whole pipeline dataset and tested on the uploaded dataset.

    """

    return sync_detailed(
        client=client,
        id_=id_,
        multipart_data=multipart_data,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    id_: int,
    multipart_data: ClassificationPredictionForm,
) -> Response[ClassificationPredictionResult]:
    kwargs = _get_kwargs(
        client=client,
        id_=id_,
        multipart_data=multipart_data,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.post(**kwargs)

    return _build_response(response=response)


async def asyncio(
    *,
    client: Client,
    id_: int,
    multipart_data: ClassificationPredictionForm,
) -> Optional[ClassificationPredictionResult]:
    """It uses features inferred from on the original pipeline dataset, but a new
    model is trained on the whole pipeline dataset and tested on the uploaded dataset.

    """

    return (
        await asyncio_detailed(
            client=client,
            id_=id_,
            multipart_data=multipart_data,
        )
    ).parsed



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.backtest_stage_config import BacktestStageConfig
from ..models.classification_backtest_config import ClassificationBacktestConfig
from ..models.classification_discovery_config import ClassificationDiscoveryConfig
from ..models.classification_specification_config import ClassificationSpecificationConfig
from ..models.feature_generation_stage_config import FeatureGenerationStageConfig
from ..models.filter_stage_config import FilterStageConfig
from ..models.lstm_backtest_stage_config import LstmBacktestStageConfig
from ..models.lstm_prediction_stage_config import LstmPredictionStageConfig
from ..models.prediction_stage_config import PredictionStageConfig
from ..models.problem_specification_config import ProblemSpecificationConfig
from ..models.refinement_stage_config import RefinementStageConfig
from ..models.stage_type import StageType
from ..models.stationarisation_stage_config import StationarisationStageConfig
from ..models.trading_simulation_stage_config import TradingSimulationStageConfig
from ..models.trading_specification_stage_config import TradingSpecificationStageConfig
from ..types import UNSET, Unset

T = TypeVar("T", bound="StageSpecification")


@attr.s(auto_attribs=True)
class StageSpecification:
    """  """

    stage_type: StageType
    config: Union[
        BacktestStageConfig,
        ClassificationBacktestConfig,
        ClassificationDiscoveryConfig,
        ClassificationSpecificationConfig,
        FeatureGenerationStageConfig,
        FilterStageConfig,
        LstmBacktestStageConfig,
        LstmPredictionStageConfig,
        PredictionStageConfig,
        ProblemSpecificationConfig,
        RefinementStageConfig,
        StationarisationStageConfig,
        TradingSimulationStageConfig,
        TradingSpecificationStageConfig,
        Unset,
    ] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        stage_type = self.stage_type.value

        config: Union[Dict[str, Any], Unset]
        if isinstance(self.config, Unset):
            config = UNSET
        elif isinstance(self.config, ProblemSpecificationConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, StationarisationStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, FeatureGenerationStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, FilterStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, RefinementStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, BacktestStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, PredictionStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, TradingSpecificationStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, TradingSimulationStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, ClassificationSpecificationConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, ClassificationDiscoveryConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, ClassificationBacktestConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        elif isinstance(self.config, LstmBacktestStageConfig):
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        else:
            config = UNSET
            if not isinstance(self.config, Unset):
                config = self.config.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "stageType": stage_type,
            }
        )
        if config is not UNSET:
            field_dict["config"] = config

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        stage_type = StageType(d.pop("stageType"))

        def _parse_config(
            data: object,
        ) -> Union[
            BacktestStageConfig,
            ClassificationBacktestConfig,
            ClassificationDiscoveryConfig,
            ClassificationSpecificationConfig,
            FeatureGenerationStageConfig,
            FilterStageConfig,
            LstmBacktestStageConfig,
            LstmPredictionStageConfig,
            PredictionStageConfig,
            ProblemSpecificationConfig,
            RefinementStageConfig,
            StationarisationStageConfig,
            TradingSimulationStageConfig,
            TradingSpecificationStageConfig,
            Unset,
        ]:
            if isinstance(data, Unset):
                return data
            try:
                config_type0: Union[Unset, ProblemSpecificationConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type0 = UNSET
                _config_type0 = data
                if not isinstance(_config_type0, Unset):
                    config_type0 = ProblemSpecificationConfig.from_dict(_config_type0)

                return config_type0
            except:  # noqa: E722
                pass
            try:
                config_type1: Union[Unset, StationarisationStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type1 = UNSET
                _config_type1 = data
                if not isinstance(_config_type1, Unset):
                    config_type1 = StationarisationStageConfig.from_dict(_config_type1)

                return config_type1
            except:  # noqa: E722
                pass
            try:
                config_type2: Union[Unset, FeatureGenerationStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type2 = UNSET
                _config_type2 = data
                if not isinstance(_config_type2, Unset):
                    config_type2 = FeatureGenerationStageConfig.from_dict(_config_type2)

                return config_type2
            except:  # noqa: E722
                pass
            try:
                config_type3: Union[Unset, FilterStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type3 = UNSET
                _config_type3 = data
                if not isinstance(_config_type3, Unset):
                    config_type3 = FilterStageConfig.from_dict(_config_type3)

                return config_type3
            except:  # noqa: E722
                pass
            try:
                config_type4: Union[Unset, RefinementStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type4 = UNSET
                _config_type4 = data
                if not isinstance(_config_type4, Unset):
                    config_type4 = RefinementStageConfig.from_dict(_config_type4)

                return config_type4
            except:  # noqa: E722
                pass
            try:
                config_type5: Union[Unset, BacktestStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type5 = UNSET
                _config_type5 = data
                if not isinstance(_config_type5, Unset):
                    config_type5 = BacktestStageConfig.from_dict(_config_type5)

                return config_type5
            except:  # noqa: E722
                pass
            try:
                config_type6: Union[Unset, PredictionStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type6 = UNSET
                _config_type6 = data
                if not isinstance(_config_type6, Unset):
                    config_type6 = PredictionStageConfig.from_dict(_config_type6)

                return config_type6
            except:  # noqa: E722
                pass
            try:
                config_type7: Union[Unset, TradingSpecificationStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type7 = UNSET
                _config_type7 = data
                if not isinstance(_config_type7, Unset):
                    config_type7 = TradingSpecificationStageConfig.from_dict(_config_type7)

                return config_type7
            except:  # noqa: E722
                pass
            try:
                config_type8: Union[Unset, TradingSimulationStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type8 = UNSET
                _config_type8 = data
                if not isinstance(_config_type8, Unset):
                    config_type8 = TradingSimulationStageConfig.from_dict(_config_type8)

                return config_type8
            except:  # noqa: E722
                pass
            try:
                config_type9: Union[Unset, ClassificationSpecificationConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type9 = UNSET
                _config_type9 = data
                if not isinstance(_config_type9, Unset):
                    config_type9 = ClassificationSpecificationConfig.from_dict(_config_type9)

                return config_type9
            except:  # noqa: E722
                pass
            try:
                config_type10: Union[Unset, ClassificationDiscoveryConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type10 = UNSET
                _config_type10 = data
                if not isinstance(_config_type10, Unset):
                    config_type10 = ClassificationDiscoveryConfig.from_dict(_config_type10)

                return config_type10
            except:  # noqa: E722
                pass
            try:
                config_type11: Union[Unset, ClassificationBacktestConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type11 = UNSET
                _config_type11 = data
                if not isinstance(_config_type11, Unset):
                    config_type11 = ClassificationBacktestConfig.from_dict(_config_type11)

                return config_type11
            except:  # noqa: E722
                pass
            try:
                config_type12: Union[Unset, LstmBacktestStageConfig]
                if not isinstance(data, dict):
                    raise TypeError()
                config_type12 = UNSET
                _config_type12 = data
                if not isinstance(_config_type12, Unset):
                    config_type12 = LstmBacktestStageConfig.from_dict(_config_type12)

                return config_type12
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            config_type13: Union[Unset, LstmPredictionStageConfig]
            config_type13 = UNSET
            _config_type13 = data
            if not isinstance(_config_type13, Unset):
                config_type13 = LstmPredictionStageConfig.from_dict(_config_type13)

            return config_type13

        config = _parse_config(d.pop("config", UNSET))

        stage_specification = cls(
            stage_type=stage_type,
            config=config,
        )

        stage_specification.additional_properties = d
        return stage_specification

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.node_and_link import NodeAndLink

T = TypeVar("T", bound="FeatureNodesAndLinks")


@attr.s(auto_attribs=True)
class FeatureNodesAndLinks:
    """ A representation of a Feature (some transformation of the original data)  """

    stage_id: int
    nodes_and_links: List[NodeAndLink]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        stage_id = self.stage_id
        nodes_and_links = []
        for nodes_and_links_item_data in self.nodes_and_links:
            nodes_and_links_item = nodes_and_links_item_data.to_dict()

            nodes_and_links.append(nodes_and_links_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "stageId": stage_id,
                "nodesAndLinks": nodes_and_links,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        stage_id = d.pop("stageId")

        nodes_and_links = []
        _nodes_and_links = d.pop("nodesAndLinks")
        for nodes_and_links_item_data in _nodes_and_links:
            nodes_and_links_item = NodeAndLink.from_dict(nodes_and_links_item_data)

            nodes_and_links.append(nodes_and_links_item)

        feature_nodes_and_links = cls(
            stage_id=stage_id,
            nodes_and_links=nodes_and_links,
        )

        feature_nodes_and_links.additional_properties = d
        return feature_nodes_and_links

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

T = TypeVar("T", bound="UserTaskStatus")


@attr.s(auto_attribs=True)
class UserTaskStatus:
    """  """

    num_fire_and_forget_tasks: int
    num_unfinished_run_tasks: int
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        num_fire_and_forget_tasks = self.num_fire_and_forget_tasks
        num_unfinished_run_tasks = self.num_unfinished_run_tasks

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "numFireAndForgetTasks": num_fire_and_forget_tasks,
                "numUnfinishedRunTasks": num_unfinished_run_tasks,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        num_fire_and_forget_tasks = d.pop("numFireAndForgetTasks")

        num_unfinished_run_tasks = d.pop("numUnfinishedRunTasks")

        user_task_status = cls(
            num_fire_and_forget_tasks=num_fire_and_forget_tasks,
            num_unfinished_run_tasks=num_unfinished_run_tasks,
        )

        user_task_status.additional_properties = d
        return user_task_status

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Optional

import httpx

from ...client import Client
from ...models.new_api_key import NewApiKey
from ...types import Response


def _get_kwargs(
    *,
    client: Client,
) -> Dict[str, Any]:
    url = "{}/apiKeys/".format(client.base_url)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
    }


def _parse_response(*, response: httpx.Response) -> Optional[NewApiKey]:
    if response.status_code == 200:
        response_200 = NewApiKey.from_dict(response.json())

        return response_200
    return None


def _build_response(*, response: httpx.Response) -> Response[NewApiKey]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    *,
    client: Client,
) -> Response[NewApiKey]:
    kwargs = _get_kwargs(
        client=client,
    )

    response = httpx.post(
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    *,
    client: Client,
) -> Optional[NewApiKey]:
    """Use this to authenticate when making API calls. Set the value of the `X-ApiKey` header to
    `keyId:keySecret`

    **WARNING:** Generating a new key will invalidate all existing keys.

    The API Key will only be provided once, and cannot be retrieved again later.
    The unhashed keySecret will never be stored in the database.
    """

    return sync_detailed(
        client=client,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
) -> Response[NewApiKey]:
    kwargs = _get_kwargs(
        client=client,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.post(**kwargs)

    return _build_response(response=response)


async def asyncio(
    *,
    client: Client,
) -> Optional[NewApiKey]:
    """Use this to authenticate when making API calls. Set the value of the `X-ApiKey` header to
    `keyId:keySecret`

    **WARNING:** Generating a new key will invalidate all existing keys.

    The API Key will only be provided once, and cannot be retrieved again later.
    The unhashed keySecret will never be stored in the database.
    """

    return (
        await asyncio_detailed(
            client=client,
        )
    ).parsed



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.algo_output_signal import AlgoOutputSignal
from ..models.algo_test_error import AlgoTestError
from ..models.algo_test_spec import AlgoTestSpec
from ..models.trading_algo import TradingAlgo

T = TypeVar("T", bound="AlgoTestResult")


@attr.s(auto_attribs=True)
class AlgoTestResult:
    """  """

    algo: TradingAlgo
    spec: AlgoTestSpec
    output: List[AlgoOutputSignal]
    stdout: str
    errors: List[AlgoTestError]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        algo = self.algo.to_dict()

        spec = self.spec.to_dict()

        output = []
        for output_item_data in self.output:
            output_item = output_item_data.to_dict()

            output.append(output_item)

        stdout = self.stdout
        errors = []
        for errors_item_data in self.errors:
            errors_item = errors_item_data.to_dict()

            errors.append(errors_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "algo": algo,
                "spec": spec,
                "output": output,
                "stdout": stdout,
                "errors": errors,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        algo = TradingAlgo.from_dict(d.pop("algo"))

        spec = AlgoTestSpec.from_dict(d.pop("spec"))

        output = []
        _output = d.pop("output")
        for output_item_data in _output:
            output_item = AlgoOutputSignal.from_dict(output_item_data)

            output.append(output_item)

        stdout = d.pop("stdout")

        errors = []
        _errors = d.pop("errors")
        for errors_item_data in _errors:
            errors_item = AlgoTestError.from_dict(errors_item_data)

            errors.append(errors_item)

        algo_test_result = cls(
            algo=algo,
            spec=spec,
            output=output,
            stdout=stdout,
            errors=errors,
        )

        algo_test_result.additional_properties = d
        return algo_test_result

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.predictions_dto import PredictionsDTO

T = TypeVar("T", bound="PredictResponse")


@attr.s(auto_attribs=True)
class PredictResponse:
    """  """

    target_original_column_name: str
    predictions: PredictionsDTO
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        target_original_column_name = self.target_original_column_name
        predictions = self.predictions.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "targetOriginalColumnName": target_original_column_name,
                "predictions": predictions,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        target_original_column_name = d.pop("targetOriginalColumnName")

        predictions = PredictionsDTO.from_dict(d.pop("predictions"))

        predict_response = cls(
            target_original_column_name=target_original_column_name,
            predictions=predictions,
        )

        predict_response.additional_properties = d
        return predict_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, cast

import attr

from ..models.feature_importance import FeatureImportance

T = TypeVar("T", bound="ClassificationPredictionResult")


@attr.s(auto_attribs=True)
class ClassificationPredictionResult:
    """ The result of a single batch of binary classification predictions  """

    desired_label: str
    dates_timestamp_milliseconds: List[int]
    prediction_pseudo_probabilities: List[float]
    predictions: List[bool]
    feature_importances: List[FeatureImportance]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        desired_label = self.desired_label
        dates_timestamp_milliseconds = self.dates_timestamp_milliseconds

        prediction_pseudo_probabilities = self.prediction_pseudo_probabilities

        predictions = self.predictions

        feature_importances = []
        for feature_importances_item_data in self.feature_importances:
            feature_importances_item = feature_importances_item_data.to_dict()

            feature_importances.append(feature_importances_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "desiredLabel": desired_label,
                "datesTimestampMilliseconds": dates_timestamp_milliseconds,
                "predictionPseudoProbabilities": prediction_pseudo_probabilities,
                "predictions": predictions,
                "featureImportances": feature_importances,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        desired_label = d.pop("desiredLabel")

        dates_timestamp_milliseconds = cast(List[int], d.pop("datesTimestampMilliseconds"))

        prediction_pseudo_probabilities = cast(List[float], d.pop("predictionPseudoProbabilities"))

        predictions = cast(List[bool], d.pop("predictions"))

        feature_importances = []
        _feature_importances = d.pop("featureImportances")
        for feature_importances_item_data in _feature_importances:
            feature_importances_item = FeatureImportance.from_dict(feature_importances_item_data)

            feature_importances.append(feature_importances_item)

        classification_prediction_result = cls(
            desired_label=desired_label,
            dates_timestamp_milliseconds=dates_timestamp_milliseconds,
            prediction_pseudo_probabilities=prediction_pseudo_probabilities,
            predictions=predictions,
            feature_importances=feature_importances,
        )

        classification_prediction_result.additional_properties = d
        return classification_prediction_result

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="OutOfDateReport")


@attr.s(auto_attribs=True)
class OutOfDateReport:
    """  """

    out_of_date: bool
    actual_data_split: Union[Unset, float] = UNSET
    intended_data_split: Union[Unset, float] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        out_of_date = self.out_of_date
        actual_data_split = self.actual_data_split
        intended_data_split = self.intended_data_split

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "outOfDate": out_of_date,
            }
        )
        if actual_data_split is not UNSET:
            field_dict["actualDataSplit"] = actual_data_split
        if intended_data_split is not UNSET:
            field_dict["intendedDataSplit"] = intended_data_split

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        out_of_date = d.pop("outOfDate")

        actual_data_split = d.pop("actualDataSplit", UNSET)

        intended_data_split = d.pop("intendedDataSplit", UNSET)

        out_of_date_report = cls(
            out_of_date=out_of_date,
            actual_data_split=actual_data_split,
            intended_data_split=intended_data_split,
        )

        out_of_date_report.additional_properties = d
        return out_of_date_report

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.trading_column_meaning import TradingColumnMeaning
from ..models.trading_specification_stage_config_algoparamvalues import (
    TradingSpecificationStageConfigAlgoparamvalues,
)
from ..types import UNSET, Unset

T = TypeVar("T", bound="TradingSpecificationStageConfig")


@attr.s(auto_attribs=True)
class TradingSpecificationStageConfig:
    """ Properties for a stage  """

    prediction_pipeline_id: int
    algo_id: int
    target_data_meaning: TradingColumnMeaning
    algo_param_values: TradingSpecificationStageConfigAlgoparamvalues
    type_: Union[Unset, str] = "trading_specification"
    prediction_pipeline_horizon: Union[Unset, int] = -1
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        prediction_pipeline_id = self.prediction_pipeline_id
        algo_id = self.algo_id
        target_data_meaning = self.target_data_meaning.value

        algo_param_values = self.algo_param_values.to_dict()

        type_ = self.type_
        prediction_pipeline_horizon = self.prediction_pipeline_horizon

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "predictionPipelineId": prediction_pipeline_id,
                "algoId": algo_id,
                "targetDataMeaning": target_data_meaning,
                "algoParamValues": algo_param_values,
            }
        )
        if type_ is not UNSET:
            field_dict["type"] = type_
        if prediction_pipeline_horizon is not UNSET:
            field_dict["predictionPipelineHorizon"] = prediction_pipeline_horizon

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        prediction_pipeline_id = d.pop("predictionPipelineId")

        algo_id = d.pop("algoId")

        target_data_meaning = TradingColumnMeaning(d.pop("targetDataMeaning"))

        algo_param_values = TradingSpecificationStageConfigAlgoparamvalues.from_dict(
            d.pop("algoParamValues")
        )

        type_ = d.pop("type", UNSET)

        prediction_pipeline_horizon = d.pop("predictionPipelineHorizon", UNSET)

        trading_specification_stage_config = cls(
            prediction_pipeline_id=prediction_pipeline_id,
            algo_id=algo_id,
            target_data_meaning=target_data_meaning,
            algo_param_values=algo_param_values,
            type_=type_,
            prediction_pipeline_horizon=prediction_pipeline_horizon,
        )

        trading_specification_stage_config.additional_properties = d
        return trading_specification_stage_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from dataclasses import dataclass

from mindfoundry.client.horizon.generated import Client

from .datasets_client import DatasetsClient
from .pipelines_client import PipelinesClient

_DEFAULT_TIMEOUT_SECONDS = 180


@dataclass
class Connection:
    base_url: str
    api_key: str


class HorizonClient:
    def __init__(
        self,
        connection: Connection,
        timeout: float = _DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        # Even though our swagger includes "servers":[{"url":"/api"}] it's not currently used by
        # openapi-python-client. See https://github.com/triaxtec/openapi-python-client/issues/112
        base_url = f"{connection.base_url}/api"
        headers = {"X-ApiKey": connection.api_key}
        client = Client(base_url=base_url, headers=headers, timeout=timeout)

        self._connection = connection
        self._datasets_client = DatasetsClient(client)
        self._pipelines_client = PipelinesClient(client)

    @property
    def datasets(self) -> DatasetsClient:
        return self._datasets_client

    @property
    def pipelines(self) -> PipelinesClient:
        return self._pipelines_client



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.trading_algo_param_spec import TradingAlgoParamSpec
from ..types import UNSET, Unset

T = TypeVar("T", bound="AlgoSpec")


@attr.s(auto_attribs=True)
class AlgoSpec:
    """ Contains the properties required to create/update an Algo  """

    code: str
    parameters_spec: List[TradingAlgoParamSpec]
    name: Union[Unset, str] = "New Algo"
    description: Union[Unset, str] = ""
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        code = self.code
        parameters_spec = []
        for parameters_spec_item_data in self.parameters_spec:
            parameters_spec_item = parameters_spec_item_data.to_dict()

            parameters_spec.append(parameters_spec_item)

        name = self.name
        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "code": code,
                "parametersSpec": parameters_spec,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        code = d.pop("code")

        parameters_spec = []
        _parameters_spec = d.pop("parametersSpec")
        for parameters_spec_item_data in _parameters_spec:
            parameters_spec_item = TradingAlgoParamSpec.from_dict(parameters_spec_item_data)

            parameters_spec.append(parameters_spec_item)

        name = d.pop("name", UNSET)

        description = d.pop("description", UNSET)

        algo_spec = cls(
            code=code,
            parameters_spec=parameters_spec,
            name=name,
            description=description,
        )

        algo_spec.additional_properties = d
        return algo_spec

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="ClassificationSpecificationConfig")


@attr.s(auto_attribs=True)
class ClassificationSpecificationConfig:
    """ Properties for a stage  """

    target_id: str
    label_to_predict: str
    active_columns: List[int]
    last_train_timestamp_millisec: Union[Unset, None, int] = UNSET
    type_: Union[Unset, str] = "classification_specification"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        target_id = self.target_id
        label_to_predict = self.label_to_predict
        active_columns = self.active_columns

        last_train_timestamp_millisec = self.last_train_timestamp_millisec
        type_ = self.type_

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "targetId": target_id,
                "labelToPredict": label_to_predict,
                "activeColumns": active_columns,
            }
        )
        if last_train_timestamp_millisec is not UNSET:
            field_dict["lastTrainTimestampMillisec"] = last_train_timestamp_millisec
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        target_id = d.pop("targetId")

        label_to_predict = d.pop("labelToPredict")

        active_columns = cast(List[int], d.pop("activeColumns"))

        last_train_timestamp_millisec = d.pop("lastTrainTimestampMillisec", UNSET)

        type_ = d.pop("type", UNSET)

        classification_specification_config = cls(
            target_id=target_id,
            label_to_predict=label_to_predict,
            active_columns=active_columns,
            last_train_timestamp_millisec=last_train_timestamp_millisec,
            type_=type_,
        )

        classification_specification_config.additional_properties = d
        return classification_specification_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.classification_score_metric import ClassificationScoreMetric

T = TypeVar("T", bound="SummaryScoreMetric")


@attr.s(auto_attribs=True)
class SummaryScoreMetric:
    """  """

    target_label: str
    type_: ClassificationScoreMetric
    value: float
    best_possible_score: float
    worst_possible_score: float
    naive_model_expected_value: float
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        target_label = self.target_label
        type_ = self.type_.value

        value = self.value
        best_possible_score = self.best_possible_score
        worst_possible_score = self.worst_possible_score
        naive_model_expected_value = self.naive_model_expected_value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "targetLabel": target_label,
                "type": type_,
                "value": value,
                "bestPossibleScore": best_possible_score,
                "worstPossibleScore": worst_possible_score,
                "naiveModelExpectedValue": naive_model_expected_value,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        target_label = d.pop("targetLabel")

        type_ = ClassificationScoreMetric(d.pop("type"))

        value = d.pop("value")

        best_possible_score = d.pop("bestPossibleScore")

        worst_possible_score = d.pop("worstPossibleScore")

        naive_model_expected_value = d.pop("naiveModelExpectedValue")

        summary_score_metric = cls(
            target_label=target_label,
            type_=type_,
            value=value,
            best_possible_score=best_possible_score,
            worst_possible_score=worst_possible_score,
            naive_model_expected_value=naive_model_expected_value,
        )

        summary_score_metric.additional_properties = d
        return summary_score_metric

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="ProblemSpecificationConfig")


@attr.s(auto_attribs=True)
class ProblemSpecificationConfig:
    """ Properties for a stage  """

    target_features: List[str]
    horizons: List[int]
    data_split: float
    active_columns: List[int]
    scale_factor_multiplier: float
    used_in_lstm: Union[Unset, bool] = UNSET
    type_: Union[Unset, str] = "problem_specification"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        target_features = self.target_features

        horizons = self.horizons

        data_split = self.data_split
        active_columns = self.active_columns

        scale_factor_multiplier = self.scale_factor_multiplier
        used_in_lstm = self.used_in_lstm
        type_ = self.type_

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "targetFeatures": target_features,
                "horizons": horizons,
                "dataSplit": data_split,
                "activeColumns": active_columns,
                "scaleFactorMultiplier": scale_factor_multiplier,
            }
        )
        if used_in_lstm is not UNSET:
            field_dict["usedInLstm"] = used_in_lstm
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        target_features = cast(List[str], d.pop("targetFeatures"))

        horizons = cast(List[int], d.pop("horizons"))

        data_split = d.pop("dataSplit")

        active_columns = cast(List[int], d.pop("activeColumns"))

        scale_factor_multiplier = d.pop("scaleFactorMultiplier")

        used_in_lstm = d.pop("usedInLstm", UNSET)

        type_ = d.pop("type", UNSET)

        problem_specification_config = cls(
            target_features=target_features,
            horizons=horizons,
            data_split=data_split,
            active_columns=active_columns,
            scale_factor_multiplier=scale_factor_multiplier,
            used_in_lstm=used_in_lstm,
            type_=type_,
        )

        problem_specification_config.additional_properties = d
        return problem_specification_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

T = TypeVar("T", bound="User")


@attr.s(auto_attribs=True)
class User:
    """ Information about a specific user  """

    id_: int
    auth_id: str
    is_support: bool
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id_ = self.id_
        auth_id = self.auth_id
        is_support = self.is_support

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id_,
                "authId": auth_id,
                "isSupport": is_support,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id_ = d.pop("id")

        auth_id = d.pop("authId")

        is_support = d.pop("isSupport")

        user = cls(
            id_=id_,
            auth_id=auth_id,
            is_support=is_support,
        )

        user.additional_properties = d
        return user

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.blueprint_stage_return_object import BlueprintStageReturnObject
from ..models.blueprint_type import BlueprintType

T = TypeVar("T", bound="BlueprintReturnObject")


@attr.s(auto_attribs=True)
class BlueprintReturnObject:
    """  """

    type_: BlueprintType
    stages: List[BlueprintStageReturnObject]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type_ = self.type_.value

        stages = []
        for stages_item_data in self.stages:
            stages_item = stages_item_data.to_dict()

            stages.append(stages_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type_,
                "stages": stages,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        type_ = BlueprintType(d.pop("type"))

        stages = []
        _stages = d.pop("stages")
        for stages_item_data in _stages:
            stages_item = BlueprintStageReturnObject.from_dict(stages_item_data)

            stages.append(stages_item)

        blueprint_return_object = cls(
            type_=type_,
            stages=stages,
        )

        blueprint_return_object.additional_properties = d
        return blueprint_return_object

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import abc
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from mindfoundry.client.horizon.client.horizon_client import HorizonClient
from mindfoundry.client.horizon.generated.models import (
    NodeAndLink,
    Pipeline,
    PredictResponse,
    StageSpecification,
    StageType,
)

ReportingCallback = Callable[[int, int, str], None]

FlexibleInputData = Union[pd.DataFrame, pd.Series, np.ndarray]


class ReportingFacility:
    def __init__(self, reporting_callback: Optional[ReportingCallback], steps_count: int):
        self._reporting_callback = reporting_callback or self._default_reporting_callback
        self._steps_count = steps_count
        self._current_step = 1

    def report(self, message: str) -> None:
        self._reporting_callback(self._current_step, self._steps_count, message)
        self._current_step = self._current_step + 1

    @staticmethod
    def _default_reporting_callback(current_step: int, max_steps: int, info: str) -> None:
        print(f"[Step {current_step}/{max_steps}]: {info}")


class HorizonModel(abc.ABC):
    def __init__(self, client: HorizonClient):
        self._client = client
        self._dataset_id: Optional[int] = None
        self._pipeline_id: Optional[int] = None

    def _add_stage(
        self,
        pipeline: Pipeline,
        stage_type: StageType,
        skip: bool,
        config: Any,
    ) -> Pipeline:
        if skip:
            return pipeline
        else:
            return self._client.pipelines.add_stage(
                pipeline_id=pipeline.summary.id_,
                spec=StageSpecification(
                    stage_type=stage_type,
                    config=config,
                ),
            )

    def _reset_model(self) -> None:
        if self._pipeline_id is not None:
            self._client.pipelines.delete(self._pipeline_id)
            self._pipeline_id = None

        if self._dataset_id is not None:
            self._client.datasets.delete(self._dataset_id)
            self._dataset_id = None

    @staticmethod
    def _make_input_data_frame(training_data: FlexibleInputData) -> pd.DataFrame:
        if isinstance(training_data, pd.DataFrame):
            return training_data

        if isinstance(training_data, pd.Series):
            return pd.DataFrame(data=[training_data])

        if isinstance(training_data, np.ndarray):
            return pd.DataFrame(data=training_data)

        raise ValueError("Training data format not supported")


class ModelPrediction:
    """ Contains all predictions obtained by calling `.predict()` on an Horizon model. """

    def __init__(
        self,
        prediction_responses_by_horizon: Dict[int, List[PredictResponse]],
        cadence_seconds: float,
        warnings: Optional[List[str]] = None,
    ):
        sorted_horizons = sorted(prediction_responses_by_horizon.keys())
        target_names = {
            r.target_original_column_name
            for prediction_responses in prediction_responses_by_horizon.values()
            for r in prediction_responses
        }

        arbitrary_horizon = sorted_horizons[0]
        arbitrary_response = prediction_responses_by_horizon[arbitrary_horizon][0]
        index_dtype = arbitrary_response.predictions.mean.index_dtype
        assert index_dtype == "datetime64[ns]", f"Datetime index not supported: {index_dtype}"

        data: Dict[str, pd.Series] = {}

        for target_name in target_names:
            time_index = [
                r.predictions.mean.index[0] + horizon * cadence_seconds * 1e9
                for horizon in sorted_horizons
                for r in prediction_responses_by_horizon[horizon]
                if r.target_original_column_name == target_name
            ]
            mean = [
                r.predictions.mean.data[0]
                for horizon in sorted_horizons
                for r in prediction_responses_by_horizon[horizon]
                if r.target_original_column_name == target_name
            ]
            lower_bound = [
                r.predictions.cb_low.data[0]
                for horizon in sorted_horizons
                for r in prediction_responses_by_horizon[horizon]
                if r.target_original_column_name == target_name
            ]
            higher_bound = [
                r.predictions.cb_high.data[0]
                for horizon in sorted_horizons
                for r in prediction_responses_by_horizon[horizon]
                if r.target_original_column_name == target_name
            ]

            time_index = pd.DatetimeIndex(data=time_index, dtype=index_dtype)
            data[f"{target_name} (mean)"] = pd.Series(data=mean, index=time_index)
            data[f"{target_name} (lower bound)"] = pd.Series(data=lower_bound, index=time_index)
            data[f"{target_name} (higher bound)"] = pd.Series(data=higher_bound, index=time_index)

        self._df = pd.DataFrame(data=data)
        self._warnings = warnings or []

    def as_df(self) -> pd.DataFrame:
        """
        Return predictions formatted as a Pandas data frame.
        :return:
            For forecasting, the format of the returned data frame is the following:
              - the number of rows is the same as the number of requested horizons. Each row
                refers to a prediction for a specific horizon in the future.
              - the index of the data frame is a DatetimeIndex. The date contained in each row
                refers to the date a given prediction is bound to occur.
              - each requested target column produces three output columns in the resulting data
                frame: one for the mean, one for the lower confidence bound and one for the
                higher confidence bound. If the target name is X, then these columns are
                respectively named "X (mean)", "X (lower bound)" and "X (higher bound)".
        """
        return self._df

    @property
    def values(self) -> np.ndarray:
        """
        Return the raw prediction values of this set of prediction as a numpy array.
        Same as calling `.as_df().values`, so please consult the documentation of `.as_df()`
        to know about the format and arrangement of rows and columns.
        """
        values: np.ndarray = self._df.values
        return values

    @property
    def warnings(self) -> List[str]:
        """ List any error that occurred during prediction. """
        return self._warnings


class FeatureExplanation:
    """ Concise summary that explains the origin of a Horizon-generated feature. """

    feature_name: str
    """ Name of the generated feature. """

    feature_id: str
    """ Id of the generated feature. """

    explanation: str
    """
    Natural language explanation of how this feature was generated and why it may be
    relevant for prediction.
    """

    formula: str
    """
    Human-readable pseudo-code representation of the transformer heuristic Horizon used
    to generate this feature.
    """

    horizon: int
    """
    Time-steps horizon to which the feature refers to. Some features may be more relevant
    for one horizon rather than another.
    """

    pearson_correlation_with_target: float
    """ Pearson correlation of the generated feature with the original target. """

    def __init__(self, node_and_link: NodeAndLink):
        self.feature_name = node_and_link.name
        self.feature_id = node_and_link.feature_id
        self.horizon = node_and_link.horizon
        self.explanation = node_and_link.transform.explanation
        self.pearson_correlation_with_target = node_and_link.pearson_correlation_with_target

        if node_and_link.transform.transform_category == "fake":
            self.formula = node_and_link.original_column_names[0]
        else:
            formula_params = node_and_link.original_column_names
            extra_params = {
                name: value
                for name, value in node_and_link.transform.params.additional_properties.items()
                if name not in {"n_parents", "name"}
            }

            for name, value in extra_params.items():
                formula_params.append(f"{name}={value}")

            self.formula = (
                f"{node_and_link.transform.transform_category}" f"({', '.join(formula_params)})"
            )

    def __str__(self) -> str:
        return self.formula


class FeatureSet:
    def __init__(
        self,
        materialized_data: Dict[int, pd.DataFrame],
        nodes_and_links: List[NodeAndLink],
    ):
        self._materialized_data = materialized_data
        self._explanations = [
            FeatureExplanation(node_and_link) for node_and_link in nodes_and_links
        ]

    def for_horizon(self, horizon: int) -> pd.DataFrame:
        """
        Return a dataframe containing the numeric values of each generated feature
        for the given horizon.

        :return:
            The returned dataframe will contain more columns than the original data.
            Each column is a new feature that Horizon found to have high predictive power.
            It may also contain less rows, since some features require aggregating more than
            one time steps. Original targets are included as well; beware they may appear as
            artificially generated features, created by stationarization.
        """
        return self._materialized_data[horizon]

    @property
    def explanations(self) -> List[FeatureExplanation]:
        """ Return a collection of feature explanations. """
        return self._explanations



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.regressor_type import RegressorType
from ..types import UNSET, Unset

T = TypeVar("T", bound="OnDemandPredictForm")


@attr.s(auto_attribs=True)
class OnDemandPredictForm:
    """  """

    horizon: int
    regressor: RegressorType
    filter_target_names: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        horizon = self.horizon
        regressor = self.regressor.value

        filter_target_names = self.filter_target_names

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "horizon": horizon,
                "regressor": regressor,
            }
        )
        if filter_target_names is not UNSET:
            field_dict["filterTargetNames"] = filter_target_names

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        horizon = d.pop("horizon")

        regressor = RegressorType(d.pop("regressor"))

        filter_target_names = d.pop("filterTargetNames", UNSET)

        on_demand_predict_form = cls(
            horizon=horizon,
            regressor=regressor,
            filter_target_names=filter_target_names,
        )

        on_demand_predict_form.additional_properties = d
        return on_demand_predict_form

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union

import attr
from dateutil.parser import isoparse

from ..models.algo_execution_data_featurehistory import AlgoExecutionDataFeaturehistory
from ..models.algo_execution_data_latesttargetallocations import (
    AlgoExecutionDataLatesttargetallocations,
)
from ..models.algo_execution_data_targetpredictions import AlgoExecutionDataTargetpredictions
from ..models.algo_execution_data_targetpredictionslowerbound import (
    AlgoExecutionDataTargetpredictionslowerbound,
)
from ..models.algo_execution_data_targetpredictionsupperbound import (
    AlgoExecutionDataTargetpredictionsupperbound,
)
from ..models.algo_execution_data_targetpricehistory import AlgoExecutionDataTargetpricehistory
from ..types import UNSET, Unset

T = TypeVar("T", bound="AlgoExecutionData")


@attr.s(auto_attribs=True)
class AlgoExecutionData:
    """All the inputs an Algo expects when executed

    Algos are executed at particular points in time, at which they are given
    recent history of all assets, predictions made by Horizon, and recent allocations.
    Predictions (and bounds) correspond to what Horizon predicts the value of each
    target will be in the future (how far in the future depends on the underlying
    pipeline making the prediction)."""

    target_price_history: Union[Unset, AlgoExecutionDataTargetpricehistory] = UNSET
    target_predictions: Union[Unset, AlgoExecutionDataTargetpredictions] = UNSET
    target_predictions_upper_bound: Union[
        Unset, AlgoExecutionDataTargetpredictionsupperbound
    ] = UNSET
    target_predictions_lower_bound: Union[
        Unset, AlgoExecutionDataTargetpredictionslowerbound
    ] = UNSET
    feature_history: Union[Unset, AlgoExecutionDataFeaturehistory] = UNSET
    dates: Union[Unset, List[datetime.datetime]] = UNSET
    latest_target_allocations: Union[Unset, AlgoExecutionDataLatesttargetallocations] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        target_price_history: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.target_price_history, Unset):
            target_price_history = self.target_price_history.to_dict()

        target_predictions: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.target_predictions, Unset):
            target_predictions = self.target_predictions.to_dict()

        target_predictions_upper_bound: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.target_predictions_upper_bound, Unset):
            target_predictions_upper_bound = self.target_predictions_upper_bound.to_dict()

        target_predictions_lower_bound: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.target_predictions_lower_bound, Unset):
            target_predictions_lower_bound = self.target_predictions_lower_bound.to_dict()

        feature_history: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.feature_history, Unset):
            feature_history = self.feature_history.to_dict()

        dates: Union[Unset, List[str]] = UNSET
        if not isinstance(self.dates, Unset):
            dates = []
            for dates_item_data in self.dates:
                dates_item = dates_item_data.isoformat()

                dates.append(dates_item)

        latest_target_allocations: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.latest_target_allocations, Unset):
            latest_target_allocations = self.latest_target_allocations.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if target_price_history is not UNSET:
            field_dict["targetPriceHistory"] = target_price_history
        if target_predictions is not UNSET:
            field_dict["targetPredictions"] = target_predictions
        if target_predictions_upper_bound is not UNSET:
            field_dict["targetPredictionsUpperBound"] = target_predictions_upper_bound
        if target_predictions_lower_bound is not UNSET:
            field_dict["targetPredictionsLowerBound"] = target_predictions_lower_bound
        if feature_history is not UNSET:
            field_dict["featureHistory"] = feature_history
        if dates is not UNSET:
            field_dict["dates"] = dates
        if latest_target_allocations is not UNSET:
            field_dict["latestTargetAllocations"] = latest_target_allocations

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        target_price_history: Union[Unset, AlgoExecutionDataTargetpricehistory] = UNSET
        _target_price_history = d.pop("targetPriceHistory", UNSET)
        if not isinstance(_target_price_history, Unset):
            target_price_history = AlgoExecutionDataTargetpricehistory.from_dict(
                _target_price_history
            )

        target_predictions: Union[Unset, AlgoExecutionDataTargetpredictions] = UNSET
        _target_predictions = d.pop("targetPredictions", UNSET)
        if not isinstance(_target_predictions, Unset):
            target_predictions = AlgoExecutionDataTargetpredictions.from_dict(_target_predictions)

        target_predictions_upper_bound: Union[
            Unset, AlgoExecutionDataTargetpredictionsupperbound
        ] = UNSET
        _target_predictions_upper_bound = d.pop("targetPredictionsUpperBound", UNSET)
        if not isinstance(_target_predictions_upper_bound, Unset):
            target_predictions_upper_bound = (
                AlgoExecutionDataTargetpredictionsupperbound.from_dict(
                    _target_predictions_upper_bound
                )
            )

        target_predictions_lower_bound: Union[
            Unset, AlgoExecutionDataTargetpredictionslowerbound
        ] = UNSET
        _target_predictions_lower_bound = d.pop("targetPredictionsLowerBound", UNSET)
        if not isinstance(_target_predictions_lower_bound, Unset):
            target_predictions_lower_bound = (
                AlgoExecutionDataTargetpredictionslowerbound.from_dict(
                    _target_predictions_lower_bound
                )
            )

        feature_history: Union[Unset, AlgoExecutionDataFeaturehistory] = UNSET
        _feature_history = d.pop("featureHistory", UNSET)
        if not isinstance(_feature_history, Unset):
            feature_history = AlgoExecutionDataFeaturehistory.from_dict(_feature_history)

        dates = []
        _dates = d.pop("dates", UNSET)
        for dates_item_data in _dates or []:
            dates_item = isoparse(dates_item_data)

            dates.append(dates_item)

        latest_target_allocations: Union[Unset, AlgoExecutionDataLatesttargetallocations] = UNSET
        _latest_target_allocations = d.pop("latestTargetAllocations", UNSET)
        if not isinstance(_latest_target_allocations, Unset):
            latest_target_allocations = AlgoExecutionDataLatesttargetallocations.from_dict(
                _latest_target_allocations
            )

        algo_execution_data = cls(
            target_price_history=target_price_history,
            target_predictions=target_predictions,
            target_predictions_upper_bound=target_predictions_upper_bound,
            target_predictions_lower_bound=target_predictions_lower_bound,
            feature_history=feature_history,
            dates=dates,
            latest_target_allocations=latest_target_allocations,
        )

        algo_execution_data.additional_properties = d
        return algo_execution_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



""" Contains all the data models used in inputs/outputs """

from .algo_execution_data import AlgoExecutionData
from .algo_execution_data_featurehistory import AlgoExecutionDataFeaturehistory
from .algo_execution_data_latesttargetallocations import AlgoExecutionDataLatesttargetallocations
from .algo_execution_data_targetpredictions import AlgoExecutionDataTargetpredictions
from .algo_execution_data_targetpredictionslowerbound import (
    AlgoExecutionDataTargetpredictionslowerbound,
)
from .algo_execution_data_targetpredictionsupperbound import (
    AlgoExecutionDataTargetpredictionsupperbound,
)
from .algo_execution_data_targetpricehistory import AlgoExecutionDataTargetpricehistory
from .algo_execution_symbol import AlgoExecutionSymbol
from .algo_execution_test_case import AlgoExecutionTestCase
from .algo_output_signal import AlgoOutputSignal
from .algo_spec import AlgoSpec
from .algo_test_error import AlgoTestError
from .algo_test_result import AlgoTestResult
from .algo_test_spec import AlgoTestSpec
from .algo_test_spec_algoparams import AlgoTestSpecAlgoparams
from .all_datasets import AllDatasets
from .backtest_stage import BacktestStage
from .backtest_stage_config import BacktestStageConfig
from .blueprint_list import BlueprintList
from .blueprint_return_object import BlueprintReturnObject
from .blueprint_stage_return_object import BlueprintStageReturnObject
from .blueprint_type import BlueprintType
from .bool_auto_update import BoolAutoUpdate
from .bool_locked import BoolLocked
from .classification_backtest_config import ClassificationBacktestConfig
from .classification_backtest_form import ClassificationBacktestForm
from .classification_backtest_insight import ClassificationBacktestInsight
from .classification_backtest_response import ClassificationBacktestResponse
from .classification_backtest_stage import ClassificationBacktestStage
from .classification_discovery_config import ClassificationDiscoveryConfig
from .classification_discovery_stage import ClassificationDiscoveryStage
from .classification_prediction_form import ClassificationPredictionForm
from .classification_prediction_result import ClassificationPredictionResult
from .classification_score_metric import ClassificationScoreMetric
from .classification_specification_config import ClassificationSpecificationConfig
from .classification_specification_stage import ClassificationSpecificationStage
from .column_passport import ColumnPassport
from .column_passport_autocorrelations import ColumnPassportAutocorrelations
from .correlation_method import CorrelationMethod
from .dataset import Dataset
from .date_value_pair import DateValuePair
from .date_value_point_schema import DateValuePointSchema
from .date_value_series import DateValueSeries
from .expert_predict_form import ExpertPredictForm
from .feature_generation_stage import FeatureGenerationStage
from .feature_generation_stage_config import FeatureGenerationStageConfig
from .feature_generator_type import FeatureGeneratorType
from .feature_importance import FeatureImportance
from .feature_nodes_and_links import FeatureNodesAndLinks
from .feature_transform import FeatureTransform
from .feature_transform_params import FeatureTransformParams
from .file_upload_form import FileUploadForm
from .filter_stage import FilterStage
from .filter_stage_config import FilterStageConfig
from .generic_data_point import GenericDataPoint
from .generic_data_series import GenericDataSeries
from .individual_dataset import IndividualDataset
from .ingestion_process import IngestionProcess
from .ingestion_status import IngestionStatus
from .json_data import JsonData
from .json_data_dtype import JsonDataDtype
from .json_data_type import JsonDataType
from .label_frequencies_grouped_by_time import LabelFrequenciesGroupedByTime
from .lstm_backtest_stage import LstmBacktestStage
from .lstm_backtest_stage_config import LstmBacktestStageConfig
from .lstm_backtest_stage_config_custommodelconfig import LstmBacktestStageConfigCustommodelconfig
from .lstm_prediction_stage import LstmPredictionStage
from .lstm_prediction_stage_config import LstmPredictionStageConfig
from .new_api_key import NewApiKey
from .node_and_link import NodeAndLink
from .ohlc_point_schema import OHLCPointSchema
from .on_demand_predict_form import OnDemandPredictForm
from .out_of_date_report import OutOfDateReport
from .pandas_series import PandasSeries
from .pipeline import Pipeline
from .pipeline_list import PipelineList
from .pipeline_specification import PipelineSpecification
from .pipeline_summary import PipelineSummary
from .predict_response import PredictResponse
from .predict_response_list import PredictResponseList
from .prediction_stage import PredictionStage
from .prediction_stage_config import PredictionStageConfig
from .predictions_dto import PredictionsDTO
from .predictions_dto_regressorimportances import PredictionsDTORegressorimportances
from .problem_specification_config import ProblemSpecificationConfig
from .problem_specification_stage import ProblemSpecificationStage
from .raw_column import RawColumn
from .rebalance_frequency import RebalanceFrequency
from .rebalance_weekly_day import RebalanceWeeklyDay
from .refinement_stage import RefinementStage
from .refinement_stage_config import RefinementStageConfig
from .regressor_type import RegressorType
from .serialiser_name import SerialiserName
from .series_overview_schema import SeriesOverviewSchema
from .stage_config_update import StageConfigUpdate
from .stage_insight_return_object import StageInsightReturnObject
from .stage_specification import StageSpecification
from .stage_status import StageStatus
from .stage_type import StageType
from .stationarisation_stage import StationarisationStage
from .stationarisation_stage_config import StationarisationStageConfig
from .stationarisation_strategy import StationarisationStrategy
from .storage_engine import StorageEngine
from .storage_specification import StorageSpecification
from .summary_score_metric import SummaryScoreMetric
from .target_transform_type import TargetTransformType
from .text_label_counts import TextLabelCounts
from .trading_algo import TradingAlgo
from .trading_algo_param_spec import TradingAlgoParamSpec
from .trading_algo_param_type import TradingAlgoParamType
from .trading_column_meaning import TradingColumnMeaning
from .trading_simulation_options import TradingSimulationOptions
from .trading_simulation_stage import TradingSimulationStage
from .trading_simulation_stage_config import TradingSimulationStageConfig
from .trading_specification_stage import TradingSpecificationStage
from .trading_specification_stage_config import TradingSpecificationStageConfig
from .trading_specification_stage_config_algoparamvalues import (
    TradingSpecificationStageConfigAlgoparamvalues,
)
from .user import User
from .user_task_status import UserTaskStatus
